/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: ClassPath.java,v 1.2 2002/03/08 21:53:19 mrobilla Exp $
 */

package at.dms.classfile;

import java.io.*;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;
import java.util.zip.ZipFile;
import java.util.zip.ZipException;
import java.util.zip.ZipEntry;

/**
 * This class implements the conceptual directory structure for .class files
 */
public class ClassPath {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs the class path.
   */
  public ClassPath(String classPath) {
    String	version = System.getProperty("java.version");
    if (System.getProperty("sun.boot.class.path") != null) {
      classPath += File.pathSeparator + System.getProperty("sun.boot.class.path");
    } else if (version.startsWith("1.2") || version.startsWith("1.3")) {
      String	home = System.getProperty("java.home");
      classPath += File.pathSeparator + home + File.separatorChar +
	"lib" + File.separatorChar + "rt.jar";
    }
    dirs = parseClassPath(classPath);
  }

  /**
   *
   */
  public static ClassDirectory[] parseClassPath(String classPath) {
    int			start, length;

    Vector	dirs  = new Vector();
    length = classPath.length();
    start  = 0;
    while (start < length) {
      int	end = classPath.indexOf(File.pathSeparatorChar, start);

      if (end == -1) {
	end = length;
      }

      try {
	File	file = new File(classPath.substring(start, end));

	if (file.isDirectory()) {
	  dirs.addElement(new DirClassDirectory(file));
	} else if (file.isFile()) {
	  // check if file is zipped (.zip or .jar)
	  if (file.getName().endsWith(".zip") || file.getName().endsWith(".jar")) {
	    try {
	      dirs.addElement(new ZipClassDirectory(new ZipFile(file)));
	    } catch (ZipException e) {
	      // it was not a zip file, ignore it
	    } catch (IOException e) {
	      // ignore it
	    }
	  }
	}
      } catch (SecurityException e) {
	// unreadable file, ignore it
      }

      start = end + 1;
    }

    ClassDirectory[]	cd = new ClassDirectory[dirs.size()];
    dirs.copyInto(cd);
    return cd;
  }

  // ----------------------------------------------------------------------
  // SINGLETON
  // ----------------------------------------------------------------------

  /**
   * initialization from a string that represents the class path
   * @param path		the classpath
   */
  public static void init(String path) {
    self = new ClassPath(path);
  }

  // ----------------------------------------------------------------------
  // LOAD CLASS
  // ----------------------------------------------------------------------

  /**
   * @return a class file that contain the class named name
   * @param name the name of the class file
   */
  public static boolean hasClassFile(String name) {
    Object	o = allLoadedFile.get(name);

    if (o != null && o != FILE_UNDEFINED) {
      return true;
    } else {
      ClassDescription file = self.getClassFile(name);

      allLoadedFile.put(name, file != null ? file : FILE_UNDEFINED);

      return file != null;
    }
  }

  /**
   * @return a class file that contain the class named name
   * @param name the name of the class file
   */
  public static ClassInfo getClassInfo(String name, boolean interfaceOnly) {
    if (!hasClassFile(name)) {
      return null;
    }

    ClassDescription clazz = (ClassDescription)allLoadedFile.get(name);
    allLoadedFile.remove(name);

    Data data = clazz.getData();
    try {
      ClassInfo ci = new ClassInfo(data.getDataInput(), interfaceOnly);
      data.release();
      return ci;
    } catch (ClassFileFormatException e) {
      data.release();
      e.printStackTrace();
      return null;
    } catch (IOException e) {
      data.release();
      e.printStackTrace();
      return null;
    }
  }

  /**
   * Returns a class file but dont read it
   */
  private ClassDescription getClassFile(String name) {
    for (int i = 0; i < dirs.length; i++) {
      ClassDescription file = dirs[i].getClassFile(name);

      if (file != null) {
	return file;
      }
    }

    return null;
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private static ClassPath	self = null;
  private static Hashtable	allLoadedFile = new Hashtable();
  private ClassDirectory[]	dirs;	// list of directories in class path
  private static ClassDescription FILE_UNDEFINED = new ClassDescription() {
    /**
     * @return an DataInput to read this class
     */
    public Data getData() {
      throw new RuntimeException("DUMMY");
    }};
}

/**
 * This class represents a conceptual directory which may hold
 * Java class files. Since Java can use archived class files found in
 * a compressed ("zip") file, this entity may or may not correspond to
 * an actual directory on disk.
 */
abstract class ClassDirectory {

  /**
   * @return a class file that contain the class named name
   * @param name the name of the class file
   */
  public abstract ClassDescription getClassFile(String name);
}

abstract class  ClassDescription {
  /**
   * @return an DataInput to read this class
   */
  public abstract Data getData();
}

class DirClassDirectory extends ClassDirectory {
  /**
   * Constructs a class directory representing a real directory
   */
  public DirClassDirectory(File dir) {
    this.dir = dir;
  }

  /**
   * @return a class file that contain the class named name
   * @param name the name of the class file
   */
  public ClassDescription getClassFile(String name) {
    name = name.replace('/', File.separatorChar) + ".class";

    final File		file = new File(dir.getPath(), name);

    if (!file.canRead()) {
      return null;
    } else {
      return new ClassDescription() {
      /**
       * @return an DataInput to read this class
       */
      public Data getData() {
	try {
	  return new Data(new FileInputStream(file));
	} catch (FileNotFoundException e) {
	  return null; // REALLY BAD (file exist in cp but is not accessible)
	}
      }};
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private File		dir;		// non null iff is a real directory
}

class ZipClassDirectory extends ClassDirectory {
  /**
   * Constructs a class directory representing a zipped file
   */
  public ZipClassDirectory(ZipFile zip) {
    this.zip = zip;
  }

  /**
   * @return a class file that contain the class named name
   * @param name the name of the class file
   */
  static int count;
  public ClassDescription getClassFile(String name) {
    name += ".class";

    final ZipEntry	entry = zip.getEntry(name);

    if (entry == null) {
      return null;
    } else {
      return new ClassDescription() {
      /**
       * @return an DataInput to read this class
       */
      public Data getData() {
	try {
	  return new Data(zip.getInputStream(entry));
	} catch (IOException e) {
	  return null; // REALLY BAD (file exist in cp but is not accessible)
	}
      }};
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private ZipFile	zip;		// non null iff is a zip or jar file
}

// optimization
class Data {
  public Data(InputStream is) {
    this.is = is;
  }

  public DataInput getDataInput() throws IOException {
    ba = getByteArray();
    int		n = 0;
    int		available;
    while (true) {
      int count = is.read(ba, n, ba.length - n);
      if (count < 0) {
	break;
      }
      available = is.available();
      n += count;
      if (n + available > ba.length) {
	byte[] temp = new byte[ba.length * 2];
	System.arraycopy(ba, 0, temp, 0, ba.length);
	ba = temp;
      } else if (available == 0) {
	break;
      }
    }

    is.close();
    is = null;

    return new DataInputStream(new ByteArrayInputStream(ba, 0, n));
  }

  public void release() {
    release(ba);
  }

  private static byte[] getByteArray() {
    if (!Constants.ENV_USE_CACHE || stack.empty()) {
      return new byte[10000];
    }
    return (byte[])stack.pop();
  }

  private static void release(byte[] arr) {
    if (Constants.ENV_USE_CACHE) {
      stack.push(arr);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private InputStream	is;
  private byte[]	ba;
  private static Stack	stack = new Stack();
}
