/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Compiler.java,v 1.2 2002/03/08 21:53:45 mrobilla Exp $
 */

package at.dms.compiler;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.util.Vector;

import at.dms.util.Message;
import at.dms.util.MessageDescription;

/**
 * This class defines the common behaviour of all KOPI compilers.
 */
public abstract class Compiler {

  /**
   * Creates a new compiler instance.
   */
  protected Compiler() {
    timestamp = System.currentTimeMillis();
  }

  /**
   * Sets the directory where to search for source files.
   * @param	dir		the name of the directory
   */
  protected void setWorkingDirectory(String dir) {
    this.workingDirectory = dir;
  }

  /**
   * Returns the timestamp of the compilation session,
   * identifying it uniquely.
   */
  public final long getTimestamp() {
    return timestamp;
  }

  /**
   * Takes a vector of file names an checks that each exists.
   * @param	files			a vector of names
   * @return	a vector of files known to exist
   * @exception	UnpositionedError	at least one file does not exist
   */
  public Vector verifyFiles(Vector names) throws UnpositionedError {
    Vector	temp = new Vector(names.size());

    // remove @files
    for (int i = 0; i < names.size(); i++) {
      if (names.elementAt(i).toString().startsWith("@")) {
	try {
	  readList(temp, workingDirectory, (String)names.elementAt(i));
	} catch (IOException e) {
	  System.err.println("Invalid list file: " + (String)names.elementAt(i));
	  System.err.println(e.getMessage());
	}
      } else {
	temp.addElement(names.elementAt(i));
      }
    }

    Vector	files = new Vector(temp.size());

    for (int i = 0; i < temp.size(); i++) {
      String	name;
      File	file;

      if (workingDirectory == null) {
	name = (String)temp.elementAt(i);
      } else {
	name = workingDirectory + File.separatorChar + (String)temp.elementAt(i);
      }
      file = new File(name);

      if (!file.exists()) {
	file = new File((String)temp.elementAt(i));
	if (!file.exists() || !file.isAbsolute()) {
	  throw new UnpositionedError(CompilerMessages.UNKNOWN_FILE, temp.elementAt(i), null);
	}
      }
      files.addElement(file);
    }

    return files;
  }

  private void readList(Vector list, String workingDirectory, String name) throws IOException {
    File		file = new File((workingDirectory == null ? "" : workingDirectory + File.separatorChar) + name.substring(1));
    LineNumberReader	reader = new LineNumberReader(new FileReader(file));
    String		str;

    while ((str = reader.readLine()) != null) {
      if (str.length() >  1) {
	list.addElement(str);
      }
    }
  }

  /**
   * Checks if destination is absolute or relative to working directory.
   */
  protected String checkDestination(String destination) {
    if (workingDirectory == null) {
      return destination;
    } else if (destination.equals("")) {
      return workingDirectory;
    } else if (new File(destination).isAbsolute()) {
      return destination;
    } else {
      return workingDirectory + File.separatorChar + destination;
    }
  }

  /**
   * Runs a compilation session
   *
   * @param	dir		the working directory
   * @param	err		the diagnostic output stream
   * @param	args		the arguments to the compiler
   * @return	true iff the compilation succeeded
   */
  public boolean run(String dir, PrintWriter err, String[] args) {
    this.workingDirectory = dir;
    this.err = err;
    
    return run(args);
  }

  // --------------------------------------------------------------------
  // METHODS TO BE IMPLEMENTED BY SUBCLASSES
  // --------------------------------------------------------------------

  /**
   * Runs a compilation session
   *
   * @param	args		the arguments to the compiler
   * @return	true iff the compilation succeeded
   */
  public abstract boolean run(String[] args);

  /**
   * Reports a trouble (error or warning).
   *
   * @param	trouble		a description of the trouble to report.
   */
  public abstract void reportTrouble(PositionedError trouble);

  /**
   * Returns true iff comments should be parsed (false if to be skipped).
   */
  public abstract boolean parseComments();

  /**
   * Returns true iff compilation runs in verbose mode.
   */
  public abstract boolean verboseMode();

  // --------------------------------------------------------------------
  // DIAGNOSTIC OUTPUT
  // --------------------------------------------------------------------

  /**
   * Write a message to the diagnostic output.
   * @param	trouble		the error caught
   */
  public void inform(Throwable trouble) {
    inform(trouble.getMessage());
  }

  /**
   * Write a message to the diagnostic output.
   * @param	trouble		the error caught
   */
  public void inform(PositionedError trouble) {
    inform(trouble.getErrorMessage());
  }

  /**
   * Write a message to the diagnostic output.
   * @param	message		the formatted message
   */
  public void inform(Message message) {
    inform(message.getMessage());
  }

  /**
   * Write a message to the diagnostic output.
   * @param	description	the message description
   * @param	parameters	the array of parameters
   */
  public void inform(MessageDescription description, Object[] parameters) {
    inform(new Message(description, parameters));
  }

  /**
   * Write a message to the diagnostic output.
   * @param	description	the message description
   * @param	parameter1	the first parameter
   * @param	parameter2	the second parameter
   */
  public void inform(MessageDescription description, Object parameter1, Object parameter2) {
    inform(description, new Object[] { parameter1, parameter2 });
  }

  /**
   * Write a message to the diagnostic output.
   * @param	description	the message description
   * @param	parameter	the parameter
   */
  public void inform(MessageDescription description, Object parameter) {
    inform(description, new Object[] { parameter });
  }

  /**
   * Write a message to the diagnostic output.
   * @param	description	the message description
   */
  public void inform(MessageDescription description) {
    inform(description, null);
  }

  /**
   * Write a text to the diagnostic output.
   * @param	message		the message text
   */
  private void inform(String message) {
    err.println(message);
    err.flush();
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  // each compilation session is identified by a timestamp
  private final long		timestamp;

  // the output stream for diagnostic messages
  private PrintWriter		err = new PrintWriter(System.err);

  // the directory where to search source files
  private String		workingDirectory;
}
