/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClassContext.java,v 1.2 2002/03/08 21:54:14 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.CWarning;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a class context during check
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CContext
 */
public class CClassContext extends CContext {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * @param	parent		the parent context or null at top level
   * @param	clazz		the corresponding clazz
   */
  CClassContext(CContext parent, CSourceClass clazz, JTypeDeclaration decl) {
    super(parent);

    this.self = clazz;
    this.decl = decl;
  }

  /**
   * Verify all final fields are initialized
   * @exception UnpositionedError	this error will be positioned soon
   */
  public void close(JTypeDeclaration decl,
		    CVariableInfo staticC,
		    CVariableInfo instanceC,
		    CVariableInfo[] constructorsC)
    throws UnpositionedError
  {
    JFieldDeclaration[]	fields = decl.getFields();

    for (int i = 0; i < fields.length; i++) {
      CSourceField	field = (CSourceField)fields[i].getField();

      if (isFieldRedefined(field.getIdent())) {
	reportTrouble(new CWarning(decl.getTokenReference(),
				   KjcMessages.FIELD_RENAME_SUPER,
				   field.getIdent(),
				   null));
      }

      if (field.isFinal()) {
	if (field.isStatic()) {
	  check(staticC != null && CVariableInfo.isInitialized(staticC.getInfo(field.getPosition())),
		KjcMessages.UNINITIALIZED_FINAL_FIELD, field.getIdent());
	} else {
	  if (!(instanceC != null && CVariableInfo.isInitialized(instanceC.getInfo(field.getPosition())))) {
	    // should be initialized in all constructors
	    for (int j = 0; j < constructorsC.length; j++) {
	      check(CVariableInfo.isInitialized(constructorsC[j].getInfo(field.getPosition())),
		    KjcMessages.UNINITIALIZED_FINAL_FIELD, field.getIdent());
	    }
	  }
	}
      } else {
	if (field.isStatic()) {
	  if (staticC != null && !CVariableInfo.isInitialized(staticC.getInfo(field.getPosition()))) {
	    reportTrouble(new CWarning(decl.getTokenReference(),
				       KjcMessages.UNINITIALIZED_FIELD,
				       field.getIdent()));	    
	  }
	} else {
	  if (!(instanceC != null && CVariableInfo.isInitialized(instanceC.getInfo(field.getPosition())))) {
	    // should be initialized in all constructors
	    for (int j = 0; j < constructorsC.length; j++) {
	      if (!CVariableInfo.isInitialized(constructorsC[j].getInfo(field.getPosition()))) {
		reportTrouble(new CWarning(decl.getTokenReference(),
					   KjcMessages.UNINITIALIZED_FIELD,
					   field.getIdent()));
		
		break;
	      }
	    }
	  }
	}
      }
    }

    //!!! graf 000123: use self.isAbstract()
    if (! self.isAbstract()) {
      // check all abstract methods are implemented
      CMethod[]	methods = self.getAbstractMethods();
      for (int i = 0; i < methods.length; i++) {
	CMethod m = self.lookupMethod(methods[i].getIdent(), methods[i].getParameters());

	check(! m.isAbstract(),
	      KjcMessages.CLASS_SHOULD_BE_ABSTRACT, self.getIdent(), m.getIdent());
      }
    }

    // check all interfaces methods are implemented
    CMethod[]	methods = self.getInterfaceMethods();
    for (int i = 0; i < methods.length; i++) {
      CMethod m = self.lookupMethod(methods[i].getIdent(), methods[i].getParameters());
      
      if (m.getIdent() != JAV_STATIC_INIT) {
	if (m.getOwner().isInterface()) {
	  if (!self.isAbstract()) {
	    check(false, KjcMessages.CLASS_SHOULD_IMPLEMENTS, self.getIdent(), m.getIdent());
	  } else {
	    // add method as an abstract method in generated class
	    // !!! TODO
	  }
	}
      }
    }
  }

  /*
   * Returns true iff a field with same name is already defined in a superclass or
   * an implemented interface.
   *
   * @param	ident		the name of the field
   */
  private boolean isFieldRedefined(String ident) throws UnpositionedError {
    try {
      // defined once in super classes ?
      return self.lookupSuperField(ident) != null;
    } catch (UnpositionedError e) {
      if (! e.hasDescription(KjcMessages.FIELD_AMBIGUOUS)) {
	throw e;
      }
      // defined more than once!
      return true;
    }
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (NEW CONTEXT)
  // ----------------------------------------------------------------------

  public CMethodContext createMethodContext(CMethod self) {
    return new CMethodContext(this, self);
  }

  public CConstructorContext createConstructorContext(CMethod self) {
    return new CConstructorContext(this, self);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Add an initializer to this context
   */
  public void addInitializer() {
    hasInitializer = true;
  }

  /**
   * Returns true if this class need initializers
   */
  public boolean hasInitializer() {
    return hasInitializer;
  }

  /**
   * getNextSyntheticIndex
   * @return an unique int value incrementaly
   */
  public int getNextSyntheticIndex() {
    return ++index;
  }

  /**
   * getNextSyntheticIndex
   * @return an unique int value incrementaly
   */
  public int getAnonymousIndex() {
    return ++anonymous;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (INFOS)
  // ----------------------------------------------------------------------

  /**
   * @param	var		the definition of a variable
   * @return	a variable info from a variable definition in current context
   */
  public int getVariableInfo(int index) {
    return ((CBodyContext)parent).getVariableInfo(index);
  } 

  /**
   * @param	var		the definition of a field
   * @return	all informations we have about this field
   */
  public int getFieldInfo(int index) {
    return fieldInfo.getInfo(index);
  }

  /**
   * @param	index		The field position in method array of local vars
   * @param	info		The information to add
   *
   * We make it a local copy of this information and at the end of this context
   * we will transfert it to the parent context according to controlFlow
   */
  public void setFieldInfo(int index, int info) {
    fieldInfo.setInfo(index, info);
  }

  /**
   *
   */
  public CVariableInfo getVariableInfo() {
    CVariableInfo tmp = fieldInfo;

    fieldInfo = new CVariableInfo(); // ==> rename it reset

    return tmp;
  }

  public void markAllFieldToInitialized() {
    for (int i = getCClass().getFieldCount() - 1; i >= 0; i--) {
      fieldInfo.setInfo(i, CVariableInfo.initialize());
    }
  }

  /**
   *
   */
  public void setVariableInfo(CVariableInfo info) {
    fieldInfo = info;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (LOOKUP)
  // ----------------------------------------------------------------------

  /**
   * lookupClass
   * search for a class with the provided type parameters
   * @param	name		method name
   * @return	the class if found, null otherwise
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CClassType lookupClass(String name) throws UnpositionedError {
    CClass	clazz = self.lookupClass(name);

    if (clazz != null) {
      return clazz.getType();
    } else {
      return super.lookupClass(name);
    }
  }

  /**
   * lookupMethod
   * search for a matching method with the provided type parameters
   * @param	ident		method name
   * @param	params		method parameters
   * @return	the method or null if not found
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CMethod lookupMethod(String ident, CType[] params) throws UnpositionedError {
    CMethod	method = self.lookupMethod(ident, params);
    if (method != null) {
      return method;
    }

    CClassContext	parent = getParentContext().getClassContext();
    if (parent != null) {
      return parent.lookupMethod(ident, params);
    } else {
      return null;
    }
  }

  /**
   * lookupField
   * @param	ident		the name of the field
   * @return	a variable from an ident in current context
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CField lookupOuterField(String ident) throws UnpositionedError {
    CClassContext	parent = getParentContext().getClassContext();

    if (parent != null) {
      CField field =  parent.lookupField(ident);

      return field != null ? field : parent.lookupOuterField(ident);
    }

    return null;
  }

  /**
   * lookupField
   * @param	ident		the name of the field
   * @return	a variable from an ident in current context
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CField lookupField(String ident) throws UnpositionedError {
    CField	field = getCClass().lookupField(ident);

    return field;
  }

  /**
   * lookupOuterLocalVariable
   * @param	ident		the name of the outer variable
   * @return	a variable from an ident in upperclass context
   */
  public JExpression lookupOuterLocalVariable(TokenReference ref, String ident) {
    if (parent instanceof CBodyContext || parent instanceof CExpressionContext) {
      JLocalVariable var = parent.lookupLocalVariable(ident);
      if (var != null) {
	return new JOuterLocalVariableExpression(ref, var, getCClass());
      } else {
	CClassContext ctxt = getParentContext().getClassContext();
	return ctxt == null ? null : ctxt.lookupOuterLocalVariable(ref, ident);
      }
    }
    return null;
  }

  /**
   * lookupLocalVariable
   * @param	ident		the name of the local variable
   * @return	a variable from an ident in current context
   */
  public JLocalVariable lookupLocalVariable(String ident) {
    return null;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (TREE HIERARCHY)
  // ----------------------------------------------------------------------

  /**
   * getClassContext
   * @return	the near parent of type CClassContext
   */
  public CClassContext getClassContext() {
    return this;
  }

  /**
   * getCClass
   * @return	the near parent of type CClassContext
   */
  public CClass getCClass() {
    return self;
  }

  /**
   * getMethod
   * @return	the near parent of type CMethodContext
   */
  public CMethodContext getMethod() {
    return getParentContext().getMethodContext();
  }

  /**
   * getTypeDeclaration
   * @return	the near parent of type CMethodContext
   */
  public JTypeDeclaration getTypeDeclaration() {
    return decl;
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CVariableInfo	fieldInfo = new CVariableInfo();
  private	JTypeDeclaration decl;
  private	boolean		hasInitializer;
  private	int		index = 0;
  private	int		anonymous = 0;
  protected	CSourceClass	self;
}
