/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CNumericType.java,v 1.2 2002/03/08 21:54:19 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.FastStringBuffer;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents java and kopi numericals types
 * Such as byte, short, int, long, float, double
 */
public class CNumericType extends CType {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructor
   * @param	typeID		the ident (int value) of this type
   */
  protected CNumericType(int typeID) {
    super(typeID);
    assert(typeID == TID_BYTE ||
	   typeID == TID_CHAR ||
	   typeID == TID_DOUBLE ||
	   typeID == TID_FLOAT ||
	   typeID == TID_INT ||
	   typeID == TID_LONG ||
	   typeID == TID_SHORT);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Transforms this type to a string
   */
  public String toString() {
    switch (type) {
    case TID_BYTE:
      return "byte";
    case TID_CHAR:
      return "char";
    case TID_DOUBLE:
      return "double";
    case TID_FLOAT:
      return "float";
    case TID_INT:
      return "int";
    case TID_LONG:
      return "long";
    case TID_SHORT:
      return "short";
    default:
      fail();
      return null;
    }
  }

  /**
   * Transforms this type to a string
   */
  public String getSignature() {
    switch (type) {
    case TID_BYTE:
      return "B";
    case TID_CHAR:
      return "C";
    case TID_DOUBLE:
      return "D";
    case TID_FLOAT:
      return "F";
    case TID_INT:
      return "I";
    case TID_LONG:
      return "J";
    case TID_SHORT:
      return "S";
    default:
      fail();
      return null;
    }
  }

  /**
   * Transforms this type to a string
   */
  protected void appendSignature(FastStringBuffer buff) {
    char	c;

    switch (type) {
    case TID_BYTE:
      c = 'B';
      break;
    case TID_CHAR:
      c = 'C';
      break;
    case TID_DOUBLE:
      c = 'D';
      break;
    case TID_FLOAT:
      c = 'F';
      break;
    case TID_INT:
      c = 'I';
      break;
    case TID_LONG:
      c = 'J';
      break;
    case TID_SHORT:
      c = 'S';
      break;
    default:
      c = '!';
      fail();
    }
    buff.append(c);
  }

  /**
   * Returns the size used in stack by value of this type
   */
  public int getSize() {
    return (type == TID_LONG || type == TID_DOUBLE) ? 2 : 1;
  }

  /**
   * Check if a type is a numeric type
   * @return is it a numeric type ?
   */
  public boolean isNumeric() {
    return true;
  }

  public boolean isCodedHasInt() {
    switch (type) {
    case TID_BYTE:
    case TID_CHAR:
    case TID_INT:
    case TID_SHORT:
      return true;
    }
    return false;
  }

  /**
   * Check if a type is an integer type
   * @return is it a integer type ?
   */
  public boolean isOrdinal() {
    switch (type) {
    case TID_BYTE:
    case TID_CHAR:
    case TID_SHORT:
    case TID_INT:
    case TID_LONG:
      return true;
    default:
      return false;
    }
  }

  /**
   * Check if a type is an integer type
   * @return is it a real number type ?
   */
  public boolean isFloatingPoint() {
    switch (type) {
    case TID_FLOAT:
    case TID_DOUBLE:
      return true;
    default:
      return false;
    }
  }

  // ----------------------------------------------------------------------
  // BODY CHECKING
  // ----------------------------------------------------------------------

  /**
   * check that type is valid
   * necessary to resolve String into java/lang/String
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkType(CContext context) throws UnpositionedError {}

  /**
   * isAssignableTo
   * @param dest  the destination type
   * @return  assignment is valid ?
   */
  public boolean isAssignableTo(CType dest) {
    if (dest == this) {
      return true;
    }
    return dest.isNumeric() && type < dest.type;
  }

  /**
   * isCastableTo
   * @param dest  the destination type
   * @return  cast is valid ?
   */
  public boolean isCastableTo(CType dest) {
    return dest.isNumeric();
  }

  /**
   * implicitlyCastTo
   * @param	dest	the expression type to cast to
   * @return	cast is valid ?
   */
  public boolean implicitlyCastTo(CType dest) {
    if (this == dest) {
      return true;
    } else if (!dest.isNumeric()) {
      return false;
    } else {
      switch (type) {
      case TID_BYTE:
	if (dest.type == TID_SHORT || dest.type == TID_CHAR) {
	  return true;
	}
      case TID_SHORT:
      case TID_CHAR:
	if (dest.type == TID_INT) {
	  return true;
	}
      case TID_INT:
	if (dest.type == TID_LONG) {
	  return true;
	}
      case TID_LONG:
	if (dest.type == TID_FLOAT) {
	  return true;
	}
      case TID_FLOAT:
	if (dest.type == TID_DOUBLE) {
	  return true;
	}
      case TID_DOUBLE:
	break;
      default:
	fail();
      }
      return false;
    }
  }

  /**
   * unifyNumeric
   * search the type corresponding to the unification of the two types
   * @param t1 the first type
   * @param t2 the second type
   * @return the corresponding type or null
   */
  public static CType unifyNumeric(CType t1, CType t2) {
    assert(t1.isNumeric() && t2.isNumeric());
    return CNumericType.unifyNumericIn(t1, CNumericType.unifyNumericIn(t2, t1));
  }

  /**
   * unifyNumeric
   * search the type corresponding to the unification of the two types
   * @param t1 the first type
   * @param t2 the second type
   * @return the corresponding type or null
   */
  private static CType unifyNumericIn(CType t1, CType t2) {
    if (t1 == CStdType.Short && t2 == CStdType.Byte) {
      return CStdType.Short;
    } else if (t1 == CStdType.Char && (t2 == CStdType.Char || t2 == CStdType.Byte)) {
      return CStdType.Char;
    } else {
      return null;
    }
  }

  /**
   * unaryPromote
   * search the type corresponding to the type after computation
   * @param t1 the type
   * @return the corresponding type after operation
   */
  public static CType unaryPromote(CType t1) {
    if (t1 == CStdType.Byte || t1 == CStdType.Short || t1 == CStdType.Char) {
      return CStdType.Integer;
    } else {
      return t1;
    }
  }

  /**
   * binaryPromote
   * search the type corresponding to the promotion of the two types
   * @param t1 the first type
   * @param t2 the second type
   * @return the corresponding type or null
   */
  public static CType binaryPromote(CType t1, CType t2) {
    assert(t1.isNumeric() && t2.isNumeric());
    if (t1 == CStdType.Double) {
      return CStdType.Double;
    } else if (t2 == CStdType.Double) {
      return CStdType.Double;
    } else if (t1 == CStdType.Float) {
      return CStdType.Float;
    } else if (t2 == CStdType.Float) {
      return CStdType.Float;
    } else if (t1 == CStdType.Long) {
      return CStdType.Long;
    } else if (t2 == CStdType.Long) {
      return CStdType.Long;
    } else {
      return CStdType.Integer;
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generates a sequence of bytescodes
   * @param	dest		the destination type
   * @param	code		the code list
   */
  public void genCastTo(CNumericType dest, CodeSequence code) {
    if (dest != this) {
      switch (dest.type) {
      case TID_BYTE:
	genCastTo(CStdType.Integer, code);
	code.plantInstruction(new NoArgInstruction(opc_i2b));
	break;

      case TID_CHAR:
	genCastTo(CStdType.Integer, code);
	code.plantInstruction(new NoArgInstruction(opc_i2c));
	break;

      case TID_SHORT:
	genCastTo(CStdType.Integer, code);
	code.plantInstruction(new NoArgInstruction(opc_i2s));
	break;

      case TID_INT:
	switch (type) {
	case TID_BYTE:
	case TID_CHAR:
	case TID_SHORT:
	  break;
	case TID_INT:
	  fail();
	case TID_LONG:
	  code.plantInstruction(new NoArgInstruction(opc_l2i));
	  break;
	case TID_FLOAT:
	  code.plantInstruction(new NoArgInstruction(opc_f2i));
	  break;
	case TID_DOUBLE:
	  code.plantInstruction(new NoArgInstruction(opc_d2i));
	  break;
	default:
	  fail();
	}
	break;

      case TID_LONG:
	switch (type) {
	case TID_BYTE:
	case TID_CHAR:
	case TID_SHORT:
	case TID_INT:
	  code.plantInstruction(new NoArgInstruction(opc_i2l));
	  break;
	case TID_LONG:
	  fail();
	case TID_FLOAT:
	  code.plantInstruction(new NoArgInstruction(opc_f2l));
	  break;
	case TID_DOUBLE:
	  code.plantInstruction(new NoArgInstruction(opc_d2l));
	  break;
	default:
	  fail();
	}
	break;
	
      case TID_FLOAT:
	switch (type) {
	case TID_BYTE:
	case TID_CHAR:
	case TID_SHORT:
	case TID_INT:
	  code.plantInstruction(new NoArgInstruction(opc_i2f));
	  break;
	case TID_LONG:
	  code.plantInstruction(new NoArgInstruction(opc_l2f));
	  break;
	case TID_FLOAT:
	  fail();
	case TID_DOUBLE:
	  code.plantInstruction(new NoArgInstruction(opc_d2f));
	  break;
	default:
	  fail();
	}
	break;
	
      case TID_DOUBLE:
	switch (type) {
	case TID_BYTE:
	case TID_CHAR:
	case TID_SHORT:
	case TID_INT:
	  code.plantInstruction(new NoArgInstruction(opc_i2d));
	  break;
	case TID_LONG:
	  code.plantInstruction(new NoArgInstruction(opc_l2d));
	  break;
	case TID_FLOAT:
	  code.plantInstruction(new NoArgInstruction(opc_f2d));
	  break;
	case TID_DOUBLE:
	  fail();
	default:
	  fail();
	}
	break;

      default:
	fail();
      }
    }
  }
}
