/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassFieldExpression.java,v 1.2 2002/03/08 21:54:26 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * a field access ('getExpression().t' or 't')
 */
public class JClassFieldExpression extends JExpression implements CInitializable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where the line of this node in the source code
   */
  public JClassFieldExpression(TokenReference where,
			       JExpression prefix,
			       String ident)
  {
    super(where);

    this.prefix = prefix;
    this.ident = ident;
  }


  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   */
  public JClassFieldExpression(TokenReference where, String ident) {
    this(where, null, ident);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  public String getIdent() {
    return ident;
  }

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return field.getType();
  }

  /**
   * @return is this expression constant ?
   */
  public boolean isConstant() {
    return field.isFinal() && field.getValue() != null && field.getValue().isConstant();
  }

  /**
   * Returns true if this field accept assignment
   */
  public boolean isLValue(CExpressionContext context) {
    if (!field.isFinal() || !(field instanceof CSourceField)) {
      return true;
    } else if (context.getBodyContext().getClassContext().getCClass() == field.getOwner()
	       && !((CSourceField)field).isFullyDeclared()) {
      return !CVariableInfo.mayBeInitialized(context.getBodyContext().getFieldInfo(((CSourceField)field).getPosition()));
    } else {
      return true;
    }
  }

  /**
   * Returns true if this field is already initialized
   */
  public boolean isInitialized(CExpressionContext context) {
    if (!(field instanceof CSourceField) || field.isStatic()) {
      return true;
    } else if (context.getBodyContext().getClassContext().getCClass() == field.getOwner() &&
	       !((CSourceField)field).isFullyDeclared()) {
      return CVariableInfo.isInitialized(context.getBodyContext().getFieldInfo(((CSourceField)field).getPosition()));
    } else {
      return true;
    }
  }

  /**
   * Returns true if this field is already initialized
   */
  public void initialize(CBodyContext context) {
    if ((field instanceof CSourceField) && (context.getClassContext().getCClass() == field.getOwner() && !((CSourceField)field).isFullyDeclared())) {
      context.setFieldInfo(((CSourceField)field).getPosition(), CVariableInfo.initialize());
    }
  }

  /**
   *
   */
  public CField getField() {
    return field;
  }

  /**
   * Returns the literal value of this field
   */
  public JLiteral getLiteral() {
    return (JLiteral)field.getValue();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    CClass	local = context.getClassContext().getCClass();

    findPrefix(local, context);
    checkAccess(local, context);

    check(context,
	  field.isStatic() || !(prefix instanceof JTypeNameExpression),
	  KjcMessages.FIELD_NOSTATIC, ident);
    if (field.isStatic() && !(prefix instanceof JTypeNameExpression)) {
      context.reportTrouble(new CWarning(getTokenReference(),
					 KjcMessages.INSTANCE_PREFIXES_STATIC_FIELD,
					 ident));
    }

    if (field instanceof CSourceField && !context.discardValue()) {
      ((CSourceField)field).setUsed();
    }

    if (isConstant()) {
      return field.getValue();
    } else {
      return this;
    }
  }

  /**
   * Find the type of the prefix
   *
   * @exception	PositionedError		Error catched as soon as possible
   */
  protected void findPrefix(CClass local, CExpressionContext context) throws PositionedError {
    if (prefix != null) {
      prefix = prefix.checkExpression(context);
      check(context,
	    prefix.getType().isClassType(),
	    KjcMessages.FIELD_BADACCESS, prefix.getType());
    } else {
      try {
	field = context.lookupField(ident);
	if (field == null) {
	  field = context.getClassContext().lookupOuterField(ident); // $$$ Why searching again
	}
      } catch (UnpositionedError e) {
	throw e.addPosition(getTokenReference());
      }
      check(context, field != null, KjcMessages.FIELD_UNKNOWN, ident);

      if (!field.isStatic()) {
	check(context,
	      !local.isStatic() || local.descendsFrom(field.getOwner()),
	      KjcMessages.FIELD_STATICERR, ident);
	prefix = new JThisExpression(getTokenReference(), field.getOwner());
      } else {
	prefix = new JTypeNameExpression(getTokenReference(), field.getOwner().getType());
      }
      prefix = prefix.checkExpression(context);
    }
  }

  /**
   * Checks is access to prefix is okay
   *
   * @exception	PositionedError	Error catched as soon as possible
   */
  public void checkAccess(CClass local, CExpressionContext context) throws PositionedError {
    CClass	access = prefix.getType().getCClass();

    try {
      field = access.lookupField(ident);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }
    check(context, field != null, KjcMessages.FIELD_UNKNOWN, ident);

    //!!!DEBUG 000213 graf (see CClassType.getCClass())
    try {
      field.getType().checkType(context);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }
    //!!!DEBUG 000213 graf
    access = field.getOwner();

    check(context, field.isAccessible(local), KjcMessages.FIELD_NOACCESS, ident);
    check(context, field.getOwner().isAccessible(local), KjcMessages.CLASS_NOACCESS, field.getOwner().getIdent());

    if ((context.getMethodContext() instanceof CConstructorContext) &&
	(prefix instanceof JThisExpression) &&
	!field.isStatic()) {
      if (field.getType().isClassType() &&
	  field.getType().getCClass() == context.getClassContext().getCClass()) {
	check(context,
	      ((CConstructorContext)context.getMethodContext()).isSuperConstructorCalled(),
	      KjcMessages.CONSTRUCTOR_EXPLICIT_CALL, field.getIdent());
      }
    }

    if (access == local) {
      /* !!!
      boolean	declared = !((field instanceof CSourceField) && !((CSourceField)field).isFullyDeclared());
      check(context, declared || (prefix != null && !(prefix instanceof JThisExpression)),
	    "field-not-declared-yet", field.getIdent());
      */
    }
  }

  public boolean equals(Object o) {
    return o instanceof JClassFieldExpression &&
      field.equals(((JClassFieldExpression)o).field) &&
      prefix.equals(((JClassFieldExpression)o).prefix);
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitFieldExpression(this, prefix, getIdent());
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    boolean		discardValue = code.discardValue();

    if (field.isStatic()) {
      if (prefix != null) {
	code.setDiscardValue(true);
	prefix.genCode(code);
	code.setDiscardValue(discardValue);
      }
    } else {
      prefix.genCode(code);
    }
    if (!discardValue) {
      field.genLoad(code);
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genStartStoreCode(CodeSequence code) {
    if (field.isStatic()) {
      if (prefix != null) {
	boolean		discardValue = code.discardValue();

	code.setDiscardValue(true);
	prefix.genCode(code);
	code.setDiscardValue(discardValue);
      }
    } else {
      prefix.genCode(code);
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genEndStoreCode(CodeSequence code, boolean discardValue) {
    if (!discardValue) {
      int	opcode;

      if (field.isStatic()) {
	if (getType().getSize() == 2) {
	  opcode = opc_dup2;
	} else {
	  opcode = opc_dup;
	}
      } else {
	if (getType().getSize() == 2) {
	  opcode = opc_dup2_x1;
	} else {
	  opcode = opc_dup_x1;
	}
      }

      code.plantInstruction(new NoArgInstruction(opcode));
    }

    field.genStore(code);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected JExpression		prefix;		// !!! graf 991205 make private
  protected String		ident;
  protected CField		field;
}
