/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JConditionalExpression.java,v 1.2 2002/03/08 21:54:27 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class implements expressions using the conditional operator ? :
 */
public class JConditionalExpression extends JExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	cond		the condition operand
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JConditionalExpression(TokenReference where,
				JExpression cond,
				JExpression left,
				JExpression right)
  {
    super(where);
    this.cond = cond;
    this.left = left;
    this.right = right;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    cond = cond.checkExpression(context);
    left = left.checkExpression(context);
    right = right.checkExpression(context);

    check(context, cond.getType() == CStdType.Boolean, KjcMessages.TRINARY_BADCOND);

    CType leftType = left.getType();
    CType rightType = right.getType();

    if (leftType.equals(rightType)) {
      type = leftType;
    } else if (leftType.isNumeric() && rightType.isNumeric()) {
      type = unifyNumeric(leftType, unifyNumeric(rightType, leftType));
      if (type == null) {
	type = CNumericType.binaryPromote(leftType, rightType);
      }
      check(context, type != null, KjcMessages.TRINARY_INCOMP, leftType, rightType);
      left = left.convertType(type, context);
      right = right.convertType(type, context);
    } else if (leftType.isReference() && rightType.isReference()) {
      if (leftType == CStdType.Null) {
	type = rightType;
      } else if (rightType == CStdType.Null) {
	type = leftType;
      } else if (leftType.isAssignableTo(rightType)) {
	type = rightType;
      } else if (rightType.isAssignableTo(leftType)) {
	type = leftType;
      } else {
	check(context, false, KjcMessages.TRINARY_INCOMP, leftType, rightType);
      }
    } else {
      check(context, false, KjcMessages.TRINARY_INCOMP, leftType, rightType);
    }

    if (cond.isConstant()) {
      if (cond.getBooleanLiteral().booleanValue()) {
	return left;
      } else {
	return right;
      }
    }

    return this;
  }

  /**
   * unifyNumeric
   * search the type corresponding to the unification of the two types
   * @param t1 the first type
   * @param t2 the second type
   * @return the corresponding type or null
   */
  private CType unifyNumeric(CType t1, CType t2) {
    if (t1 == CStdType.Short && t2 == CStdType.Byte) {
      return CStdType.Short;
    } else if (t2 == CStdType.Short && t1 == CStdType.Byte) {
      return CStdType.Short;
    } else if (t1 == CStdType.Char && (t2 == CStdType.Char || t2 == CStdType.Byte)) {
      return CStdType.Char;
    } else if (t2 == CStdType.Char && (t1 == CStdType.Char || t1 == CStdType.Byte)) {
      return CStdType.Char;
    } else {
      return null;
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitConditionalExpression(this, cond, left, right);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    CodeLabel		rightLabel = new CodeLabel();
    CodeLabel		nextLabel = new CodeLabel();

    cond.genBranch(false, code, rightLabel);		//		COND IFEQ right
    left.genCode(code); 			//		LEFT CODE
    code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
    code.plantLabel(rightLabel);		//	right:
    right.genCode(code);			//		RIGHT CODE
    code.plantLabel(nextLabel);			//	next	...
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private CType		type;
  private JExpression   cond;
  private JExpression   left;
  private JExpression   right;
}
