/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JEqualityExpression.java,v 1.2 2002/03/08 21:54:29 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * JLS 15.20: Equality Operators ('==' and '!=')
 */
public class JEqualityExpression extends JBinaryExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	oper		the operator
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JEqualityExpression(TokenReference where,
			     int oper,
			     JExpression left,
			     JExpression right)
  {
    super(where, left, right);
    this.oper = oper;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    left = left.checkExpression(context);
    right = right.checkExpression(context);

    CType	leftType = left.getType();
    CType	rightType = right.getType();

    if (leftType.isNumeric()) {
      // JLS 15.20.1: Numerical Equality Operators
      check(context, rightType.isNumeric(), KjcMessages.EQUALITY_TYPE, leftType, rightType);
      promote(context);
    } else if (leftType == CStdType.Boolean) {
      // JLS 15.20.2: Boolean Equality Operators
      check(context, rightType == CStdType.Boolean, KjcMessages.EQUALITY_TYPE, leftType, rightType);
      if (left instanceof JBooleanLiteral || right instanceof JBooleanLiteral) {
	context.reportTrouble(new CWarning(getTokenReference(),
					   KjcMessages.COMPARING_BOOLEAN_CONSTANT));            
      }
    } else {
      // JLS 15.20.3: Reference Equality Operators
      check(context,
	    leftType.isReference() && rightType.isReference(),
	    KjcMessages.EQUALITY_TYPE, leftType, rightType);
      check(context,
	    rightType.isCastableTo(leftType) || leftType.isCastableTo(rightType),
	    KjcMessages.EQUALITY_TYPE, leftType, rightType);
      if (left.getType().equals(CStdType.String) && right.getType().equals(CStdType.String) &&
	  (left.isConstant() || right.isConstant())) {
	context.reportTrouble(new CWarning(getTokenReference(),
					   KjcMessages.STRING_COMPARISON));      
      }
    }

    type = CStdType.Boolean;

    if (left.isConstant() && right.isConstant()) {
      return constantFolding(left.getLiteral(), right.getLiteral());
    } else {
      return this;
    }
  }

  /**
   * @return	a literal resulting of an operation over two literals
   */
  public JExpression constantFolding(JLiteral left, JLiteral right) {
    if (left.getType().isNumeric()) {
      if (oper == OPE_EQ) {
	return new JBooleanLiteral(getTokenReference(),
				   left.getNumberLiteral().numberValue().doubleValue() ==
				   right.getNumberLiteral().numberValue().doubleValue());
      } else {
	return new JBooleanLiteral(getTokenReference(),
				   left.getNumberLiteral().numberValue().doubleValue() !=
				   right.getNumberLiteral().numberValue().doubleValue());
      }
    } else if (left.getType() == CStdType.Boolean) {
      if (oper == OPE_EQ) {
	return new JBooleanLiteral(getTokenReference(),
				   left.getBooleanLiteral().booleanValue() ==
				   right.getBooleanLiteral().booleanValue());
      } else {
	return new JBooleanLiteral(getTokenReference(),
				   left.getBooleanLiteral().booleanValue() !=
				   right.getBooleanLiteral().booleanValue());
      }
    } else if (left instanceof JNullLiteral && right instanceof JNullLiteral) {
      if (oper == OPE_EQ) {
	return new JBooleanLiteral(getTokenReference(), true);
      } else {
	return new JBooleanLiteral(getTokenReference(), false);
      }
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitEqualityExpression(this, oper, left, right);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    genBooleanResultCode(code);
  }

  /**
   * Optimize a bi-conditional expression
   */
  protected void genBranch(JExpression left,
			   JExpression right,
			   boolean cond,
			   CodeSequence code,
			   CodeLabel label)
  {
    code.setLineNumber(getTokenReference().getLine());

    if ((left.isConstant() || right.isConstant())) {
      if (left instanceof JNullLiteral || right instanceof JNullLiteral) {
	if (left instanceof JNullLiteral) {
	  JExpression temp = left;
	  left = right;
	  right = temp;
	}
	left.genCode(code);
	code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_ifnull : opc_ifnonnull, label);
	return;
      } else if (((left instanceof JOrdinalLiteral && ((JOrdinalLiteral)left).numberValue().intValue() == 0) ||
		 (right instanceof JOrdinalLiteral && ((JOrdinalLiteral)right).numberValue().intValue() == 0)) &&
		 left.getType().isCodedHasInt()) {
	if ((left instanceof JOrdinalLiteral) &&((JOrdinalLiteral)left).numberValue().intValue() == 0) {
	  JExpression temp = left;
	  left = right;
	  right = temp;
	}
	left.genCode(code);
	code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_ifeq : opc_ifne, label);
	return;
      }
    }

    left.genCode(code);
    right.genCode(code);

    switch (left.getType().getTypeID()) {
    case TID_ARRAY:
    case TID_CLASS:
      code.plantLabelRef(cond  == (oper == OPE_EQ) ? opc_if_acmpeq : opc_if_acmpne, label);
      return;
    case TID_FLOAT:
      code.plantInstruction(new NoArgInstruction(opc_fcmpl));
      break;
    case TID_LONG:
      code.plantInstruction(new NoArgInstruction(opc_lcmp));
      break;
    case TID_DOUBLE:
      code.plantInstruction(new NoArgInstruction(opc_dcmpl));
      break;
    default:
      code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_if_icmpeq : opc_if_icmpne, label);
      return;
    }

    code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_ifeq : opc_ifne, label);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected final int			oper;
}

