/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JExpression.java,v 1.2 2002/03/08 21:54:29 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.util.MessageDescription;

/**
 * Root class for all expressions
 */
public abstract class JExpression extends JPhylum {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------
  
  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   */
  public JExpression(TokenReference where) {
    super(where);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Returns the type of this expression (call after parsing only)
   */
  public abstract CType getType();

  /**
   * @return true if this expression is constant
   */
  public boolean isConstant() {
    return false;
  }

  /**
   * Returns true iff this expression can be used as a statement (JLS 14.7)
   */
  public boolean isStatementExpression() {
    return false;
  }

  // ----------------------------------------------------------------------
  // LITERAL HANDLING
  // ----------------------------------------------------------------------

  /**
   * Returns the literal value of this expression
   */
  public JLiteral getLiteral() {
    throw new RuntimeException("THIS IS NOT A LITERAL");
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JOrdinalLiteral getOrdinalLiteral() {
    return (JOrdinalLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JRealLiteral getRealLiteral() {
    return (JRealLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JBooleanLiteral getBooleanLiteral() {
    return (JBooleanLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JNumberLiteral getNumberLiteral() {
    return (JNumberLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JStringLiteral getStringLiteral() {
    return (JStringLiteral)getLiteral();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public abstract JExpression checkExpression(CExpressionContext context) throws PositionedError;

  /**
   * Check expression and evaluate and alter context
   * @param	context		the actual context of analyse
   * @param	isLeftSide
   * @param	discardValue
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public final JExpression checkExpression(CExpressionContext context,
					   boolean isLeftSide,
					   boolean discardValue)
    throws PositionedError
  {
    boolean	isLeftSideSaved = context.isLeftSide();
    boolean	discardValueSaved = context.discardValue();

    try {
      context.setLeftSide(isLeftSide);
      context.setDiscardValue(discardValue);
      return checkExpression(context);
    } finally {
      context.setLeftSide(isLeftSideSaved);
      context.setDiscardValue(discardValueSaved);
    }
  }


  /**
   * Adds a compiler error.
   * @param	context		the context in which the error occurred
   * @param	key		the message ident to be displayed
   * @param	params		the array of parameters
   *
   */
  protected void fail(CContext context, MessageDescription key, Object[] params)
    throws PositionedError
  {
    throw new CLineError(getTokenReference(), key, params);
  }

  /**
   * convertType
   * changes the type of this expression to an other
   * @param  dest the destination type
   */
  public JExpression convertType(CType dest, CExpressionContext context) throws PositionedError {
    if (getType().equals(dest)) {
      return this;
    } else if (isConstant() && getType().isNumeric()) {
      // not null
      return JLiteral.createLiteral(dest, this.getLiteral().getValue());
    } else {
      return new JUnaryPromote(this, dest).checkExpression(context);
    }
  }

  /**
   * Can this expression be converted to the specified type by assignment conversion (JLS 5.2) ?
   * @param	dest		the destination type
   * @return	true iff the casting conversion is valid
   */
  public boolean isAssignableTo(CType dest) {
    return getType().isAssignableTo(dest);
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public abstract void accept(KjcVisitor p);

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public abstract void genCode(CodeSequence code);

  /**
   * Generates a sequence of bytescodes to branch on a label
   * This method helps to handle heavy optimizables conditions
   * @param	code		the code list
   */
  public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
    genCode(code);
    code.plantLabelRef(cond ? opc_ifne : opc_ifeq, label);
  }

  /**
   * Generates a sequence of bytescodes to branch on a label
   * This method helps to handle heavy optimizables conditions
   * @param	code		the code list
   */
  public JExpression unParenthesize() {
    return this;
  }

  // ----------------------------------------------------------------------
  // PUBLIC CONSTANTS
  // ----------------------------------------------------------------------

  public static final JExpression[]		EMPTY = new JExpression[0];
}
