/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JFieldDeclaration.java,v 1.2 2002/03/08 21:54:29 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents a java class in the syntax tree
 */
public class JFieldDeclaration extends JMemberDeclaration {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	parent		the owner of this field
   * @param	variable	the variable definition
   * @param	javadoc	is this field deprecated
   */
  public JFieldDeclaration(TokenReference where,
			   JVariableDefinition variable,
			   JavadocComment javadoc,
			   JavaStyleComment[] comments) {
    super(where, javadoc, comments);

    this.variable = variable;
  }

  // ----------------------------------------------------------------------
  // ACCESSOR
  // ----------------------------------------------------------------------

  /**
   * Returns true if this field declarator has initializer (should be initialized)
   */
  public boolean hasInitializer() {
    return variable.hasInitializer();
  }

  /**
   * Returns the type of this field
   */
  public CType getType() {
    return variable.getType();
  }

  /**
   * Returns true if this field need to be initialized
   * WARNING: this method return true when initial value corresponds to a default value
   * ====> a second check should be made after checkStatement to ensure that an initialization
   * is really needed
   */
  public boolean needInitialization() {
    if (!hasInitializer()) {
      return false;
    }

    return true; //(!variable.getValue().isConstant() || getType().isReference() || isDefaultValue());
  }

  public JVariableDefinition getVariable() {
    return variable;
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

    /**
     * Second pass (quick), check interface looks good
     * Exceptions are not allowed here, this pass is just a tuning
     * pass in order to create informations about exported elements
     * such as Classes, Interfaces, Methods, Constructors and Fields
     *  sub classes must check modifiers and call checkInterface(super)
     * @param v a vector to collect fields
     * @return true iff sub tree is correct enought to check code
     * @exception	PositionedError	an error with reference to the source file
     */
    public CSourceField checkInterface(CClassContext context) throws PositionedError
    {
	int	modifiers = variable.getModifiers();

	// Check modifiers
	// 8.3.1 Field Modifiers
	check(context,
	      !hasOtherFlags(modifiers,
			     ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE + ACC_FINAL +
			     ACC_STATIC + ACC_TRANSIENT + ACC_VOLATILE),
	      KjcMessages.FIELD_FLAGS);
	
	if (context.getCClass().isInterface())
	{
	    modifiers |= ACC_PUBLIC | ACC_STATIC | ACC_FINAL;
	    check(context,
		  !hasOtherFlags(modifiers, ACC_PUBLIC + ACC_FINAL + ACC_STATIC),
		  KjcMessages.FIELD_FLAGS_IN_INTERFACE);
	}

	variable.checkInterface(context);
	setInterface(new CSourceField(context.getCClass(),
				      modifiers,
				      variable.getIdent(),
				      variable.getType(),
				      isDeprecated()));
	
	/*
	  if (variable.getIdent() == JAV_OUTER_THIS) {
	  // remove somewhere else vgp
	  ((CSourceField)getField()).setFullyDeclared(true);
	  }
	*/
	return (CSourceField)getField();
    }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception PositionedError Error catched as soon as possible
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    variable.checkBody(context);

    if (hasInitializer() && getField().isFinal()) {
      JExpression	value = variable.getValue();

      if (value.isConstant()) {
	getField().setValue(value.getLiteral());

	if (! getField().isStatic()) {
	  context.reportTrouble(new CWarning(getTokenReference(),
					     KjcMessages.FINAL_FIELD_IMPLICITLY_STATIC,
					     getField().getIdent()));
	}
      }
    }
    
    if (hasInitializer()) {
      context.setFieldInfo(((CSourceField)getField()).getPosition(), CVariableInfo.initialize());
    }

    if (! (getField().isPublic() || getField().isPrivate())) {
      context.reportTrouble(new CWarning(getTokenReference(),
					 KjcMessages.PACKAGE_PROTECTED_ATTRIBUTE,
					 getField().getIdent()));
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);
    p.visitFieldDeclaration(this,
			    variable.getModifiers(),
			    variable.getType(),
			    variable.getIdent(),
			    variable.getValue());
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    if (variable.getValue() != null) {
      code.setLineNumber(getTokenReference().getLine());

      if (!getField().isStatic()) {
	code.plantLoadThis();
      }
      variable.getValue().genCode(code);
      getField().genStore(code);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected JVariableDefinition		variable;
}
