/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMemberDeclaration.java,v 1.2 2002/03/08 21:54:32 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.TokenReference;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.JavaStyleComment;


/**
 * This class represents a java class in the syntax tree
 */
public abstract class JMemberDeclaration extends JPhylum {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	modifiers	list of modifiers
   * @param	ident		the name of this method
   * @param	javadoc	is this member deprecated
   */
  public JMemberDeclaration(TokenReference where,
			    JavadocComment javadoc,
			    JavaStyleComment[] comments) {
    super(where);
    this.comments = comments;
    this.javadoc = javadoc;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (INTERFACE)
  // ----------------------------------------------------------------------

  /**
   * Returns true if this member is deprecated
   */
  public boolean isDeprecated() {
    return javadoc != null && javadoc.isDeprecated();
  }

  /**
   * @return	the interface
   */
  public CField getField() {
    return export.getField();
  }

  /**
   * @return	the interface
   */
  public CMethod getMethod() {
    return export.getMethod();
  }

  /**
   * @return	the interface
   */
  public CClass getCClass() {
    return export.getCClass();
  }

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    genComments(p);
  }

  /**
   * Generate the code in pure java form
   * It is useful to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genComments(KjcVisitor p) {
    if (comments != null) {
      p.visitComments(comments);
    }
    if (javadoc != null) {
      p.visitJavadoc(javadoc);
    }
  }

  // ----------------------------------------------------------------------
  // STATIC UTILITIES
  // ----------------------------------------------------------------------

  public static boolean checkOrder(int currentModifiers, int newModifier) {
    return getMaxPositionFromModifiers(currentModifiers) < getModifierPosition(newModifier);
  }

  private static int getMaxPositionFromModifiers(int mod) {
    int		max = 0;
    
    for (int i = 0; i < ACC_NBR_FLAGS; i++) {
      if (((1 << i) & mod) != 0) {
	max = Math.max(max, getModifierPosition(1 << i));
      }
    }

    return max;
  }

  private static int getModifierPosition(int mod) {
    switch (mod) {
    case ACC_PUBLIC:
      return 1;
    case ACC_PRIVATE:
      return 2;
    case ACC_PROTECTED:
      return 3;
    case ACC_ABSTRACT:
      return 4;
    case ACC_STATIC:
      return 5;
    case ACC_FINAL:
      return 6;
    case ACC_SYNCHRONIZED:
      return 7;
    case ACC_TRANSIENT:
      return 8;
    case ACC_VOLATILE:
      return 9;
    case ACC_NATIVE:
      return 10;
    case ACC_INTERFACE:
      return 11;
    default:
      throw new RuntimeException();
    }
  }

  public static String getModifierName(int mod) {
    switch (mod) {
    case ACC_PUBLIC:
      return "public";
    case ACC_PRIVATE:
      return "private";
    case ACC_PROTECTED:
      return "protected";
    case ACC_STATIC:
      return "static";
    case ACC_FINAL:
      return "final";
    case ACC_SYNCHRONIZED:
      return "synchronized";
    case ACC_VOLATILE:
      return "volatile";
    case ACC_TRANSIENT:
      return "transient";
    case ACC_NATIVE:
      return "native";
    case ACC_INTERFACE:
      return "interface";
    case ACC_ABSTRACT:
      return "abstract";
    default:
      throw new RuntimeException();
    }
  }

  // ----------------------------------------------------------------------
  // PROTECTED ACCESSORS
  // ----------------------------------------------------------------------

  protected void setInterface(CMember export) {
    this.export = export;
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CMember				export;
  private	JavadocComment			javadoc;
  private	JavaStyleComment[]		comments;
}
