/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodCallExpression.java,v 1.2 2002/03/08 21:54:32 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents methods call
 * <pre> methodname(explist) </pre>
 */
public class JMethodCallExpression extends JExpression {

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	prefix		an expression that is a field of a class representing a method
   * @param	ident		the method ident
   * @param	args		the argument of the call
   */
  public JMethodCallExpression(TokenReference where,
			       JExpression prefix,
			       JExpression[] args)
  {
    super(where);

    this.prefix = ((JNameExpression)prefix).getPrefix();
    this.ident = ((JNameExpression)prefix).getName().intern();
    this.args = args;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return the type of this expression
   */
  public CType getType() {
    return method.getReturnType();
  }

  /**
   * Returns true iff this expression can be used as a statement (JLS 14.7)
   */
  public boolean isStatementExpression() {
    return true;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    CType[]	argTypes = new CType[args.length];

    for (int i = 0; i < argTypes.length; i++) {
      args[i] = args[i].checkExpression(context, false, false);
      argTypes[i] = args[i].getType();
    }

    if (ident == null) {
      if (prefix instanceof JClassFieldExpression) {
	// condition as if-statement because of arguments to method check
	check(context,
	      false,
	      KjcMessages.FIELD_AS_METHOD, ((JClassFieldExpression)prefix).getIdent());
      }
      if (prefix instanceof JTypeNameExpression) {
	// condition as if-statement because of arguments to method check
	check(context,
	      false,
	      KjcMessages.TYPE_AS_METHOD, ((JTypeNameExpression)prefix).getType());
      }
      check(context, false, KjcMessages.BAD_METHOD_NAME, prefix);
    }

    if (prefix != null) {
      prefix = prefix.checkExpression(context, false, false);
      if (prefix instanceof JNameExpression) {
	// condition as if-statement because of arguments to method check
	check(context,
	      false,
	      KjcMessages.BAD_METHOD_NAME, ((JNameExpression)prefix).getName());
      }
      check(context,
	    prefix.getType().isReference(),
	    KjcMessages.METHOD_BADPREFIX, ident, prefix.getType());
      try {
	method = prefix.getType().getCClass().lookupMethod(ident, argTypes);
      } catch (UnpositionedError e) {
	throw e.addPosition(getTokenReference());
      }
    } else {
      try {
	method = context.lookupMethod(ident, argTypes);
      } catch (UnpositionedError e) {
	throw e.addPosition(getTokenReference());
      }
    }

    if (method == null) {
      String	prefixName;

      if (prefix instanceof JNameExpression) {
	prefixName = ((JNameExpression)prefix).getQualifiedName() + ".";
      } else if (prefix instanceof JTypeNameExpression) {
	prefixName =  ((JTypeNameExpression)prefix).getQualifiedName() + ".";
      } else {
	prefixName = "";
      }

      throw new CMethodNotFoundError(getTokenReference(), this, prefixName + ident, argTypes);
    }

    // check access
    CClass		local = context.getClassContext().getCClass();
    CClass		access = method.getOwner();

    check(context, method.isAccessible(local), KjcMessages.METHOD_NOACCESS, method.getIdent());
    //check(context, access.isAccessible(local), KjcMessages.CLASS_NOACCESS, access.getIdent());

    CClassType[]	exceptions = method.getThrowables();
    for (int i = 0; i < exceptions.length; i++) {
      if (exceptions[i].isCheckedException()) {
	if (prefix == null ||				// special case of clone
	    !prefix.getType().isArrayType() ||
	    ident != Constants.JAV_CLONE ||
	    !exceptions[i].getCClass().getQualifiedName().
	    equals("java/lang/CloneNotSupportedException")) {
	  context.getBodyContext().addThrowable(new CThrowableInfo(exceptions[i],
								   this,
								   context.getBodyContext().makeSnapshot()));
	}
      }
    }

    if ((prefix == null) && !method.isStatic()) {
      if (access == local) {
	prefix = new JThisExpression(getTokenReference()).checkExpression(context);
      } else {
	prefix = new JThisExpression(getTokenReference(), access).checkExpression(context);
      }
    }

    check(context,
	  method.isStatic() || !(prefix instanceof JTypeNameExpression),
	  KjcMessages.METHOD_STATIC_BAD, method.getIdent());

    if (method.isStatic() && prefix != null && !(prefix instanceof JTypeNameExpression)) {
      context.reportTrouble(new CWarning(getTokenReference(),
					 KjcMessages.INSTANCE_PREFIXES_STATIC_METHOD,
					 method.getIdent()));
    }

    argTypes = method.getParameters();
    for (int i = 0; i < argTypes.length; i++) {
      if (args[i] instanceof JTypeNameExpression) {
	check(context, false, KjcMessages.VAR_UNKNOWN, ((JTypeNameExpression)args[i]).getQualifiedName());
      }
      args[i] = args[i].convertType(argTypes[i], context);
    }

    // Mark method as used if it is a source method
    if (method instanceof CSourceMethod) {
      ((CSourceMethod)method).setUsed();
    }

    /** !!! graf 991210 disabled, but it should be possible to remove warning with a special tag
    if (method.getReturnType() != CStdType.Void && context.discardValue()) {
      context.reportTrouble(new CWarning(getTokenReference(),
					 KjcMessages.UNUSED_RETURN_VALUE_FROM_FUNCTION_CALL,
					 method.getIdent()));
					 
			 
    }
    */

    //String lOwner = method.getOwner().getType().toString();
    //System.out.println( lOwner + "." + method.getIdent() + method.getSignature() );

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    if (ident == null) {
      ident = ((JNameExpression)prefix).getName();
      prefix = ((JNameExpression)prefix).getPrefix();
    }
    p.visitMethodCallExpression(this, prefix, ident, args);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    boolean		forceNonVirtual = false;
    boolean		discardValue = code.discardValue();

    code.setDiscardValue(false);

    if (prefix != null) {
      if (method.isStatic()) {
	code.setDiscardValue(true);
      }
      prefix.genCode(code);

      if (prefix instanceof JSuperExpression) {
	forceNonVirtual = true;
      }
    }

    code.setDiscardValue(false);

    for (int i = 0; i < args.length; i++) {
      args[i].genCode(code);
    }

    method.genCode(code, forceNonVirtual);

    if (discardValue) {
      if (getType().getSize() == 2) {
	code.plantInstruction(new NoArgInstruction(opc_pop2));
      } else if (getType().getSize() == 1) {
	code.plantInstruction(new NoArgInstruction(opc_pop));
      }
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected JExpression		prefix;
  protected String		ident;
  protected JExpression[]	args;

  protected CMethod		method;
}
