/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodDeclaration.java,v 1.2 2002/03/08 21:54:32 mrobilla Exp $
 *
 * Modified by Martin P. Robillard for Jex 26/06/2001
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a java class in the syntax tree
 */
public class JMethodDeclaration extends JMemberDeclaration {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	parent		parent in which this methodclass is built
   * @param	modifiers	list of modifiers
   * @param	returnType	the return type of this method
   * @param	ident		the name of this method
   * @param	parameters	the parameters of this method
   * @param	exceptions	the exceptions throw by this method
   * @param	body		the body of this method
   * @param	javadoc		is this method deprecated
   */
  public JMethodDeclaration(TokenReference where,
			    int modifiers,
			    CType returnType,
			    String ident,
			    JFormalParameter[] parameters,
			    CClassType[] exceptions,
			    JBlock body,
			    JavadocComment javadoc,
			    JavaStyleComment[] comments) {
    super(where, javadoc, comments);

    this.modifiers = modifiers;
    this.returnType = returnType;
    this.ident = ident.intern();
    this.body = body;

    this.parameters = parameters;
    this.exceptions = exceptions;
    assert(parameters != null);
    assert(exceptions != null);
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------
    
    /**
     * Second pass (quick), check interface looks good
     * Exceptions are not allowed here, this pass is just a tuning
     * pass in order to create informations about exported elements
     * such as Classes, Interfaces, Methods, Constructors and Fields
     * @return true iff sub tree is correct enough to check code
     * @exception	PositionedError	an error with reference to the source file
     */
    public CSourceMethod checkInterface(CClassContext context) throws PositionedError
    {
	boolean	inInterface = context.getCClass().isInterface();
	boolean	isExported = !(this instanceof JInitializerDeclaration);
	String	ident = (this instanceof JConstructorDeclaration) ? JAV_CONSTRUCTOR : this.ident;

	// Collect all parsed data
	if (inInterface && isExported)
	{
	    modifiers |= ACC_PUBLIC | ACC_ABSTRACT;
	}

	// 8.4.3 Method Modifiers
	check(context,
	      !hasOtherFlags(modifiers,
			     ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE +
			     ACC_ABSTRACT + ACC_FINAL + ACC_STATIC +
			     ACC_NATIVE + ACC_SYNCHRONIZED),
	      KjcMessages.METHOD_FLAGS);
	
	if (inInterface && isExported)
	{
	    check(context,
		  !hasOtherFlags(modifiers, ACC_PUBLIC + ACC_ABSTRACT),
		  KjcMessages.METHOD_FLAGS_IN_INTERFACE, this.ident);
	}

	try
	{
	    returnType.checkType(context);

	    CType[]	parameterTypes = new CType[parameters.length];
	    for (int i = 0; i < parameterTypes.length; i++)
	    {
		parameterTypes[i] = parameters[i].checkInterface(context);
	    }

	    for (int i = 0; i < exceptions.length; i++)
	    {
		exceptions[i].checkType(context);
	    }

	    int	optimizeCode = (context.getCompiler() instanceof Main) ?
		((Main)context.getCompiler()).optimizeCode() :// It's a normal compiler, look in the options
		0;

	    setInterface(new CSourceMethod(context.getCClass(), modifiers, ident, returnType,
					   parameterTypes, exceptions, isDeprecated(), body,
					   optimizeCode));
      
	    return (CSourceMethod)getMethod();
	}
	catch (UnpositionedError cue)
	{
	    throw cue.addPosition(getTokenReference());
	}
    }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception PositionedError Error catched as soon as possible
   */
  public void checkStatement(CClassContext context) throws PositionedError {
    try {
      check(context,
	    context.getCClass().isAbstract() || !getMethod().isAbstract(),
	    KjcMessages.METHOD_ABSTRACT_CLASSNOT, ident);

      CMethod superMethod = context.getCClass().lookupSuperMethod(ident, getMethod().getParameters());
      
      if (superMethod != null) {
	checkOverriding(context, superMethod);
      }

      if (body != null) {
	checkBody(context);

	CMethodContext	self = new CMethodContext(context, getMethod());
	CBlockContext	block = new CBlockContext(self, parameters.length);
	CBodyContext	inside = new CBodyContext(block);

	if (!getMethod().isStatic()) {
	  // add this local var
	  block.addThisVariable();
	}

	for (int i = 0; i < parameters.length; i++) {
	  parameters[i].checkBody(inside);
	}

	body.checkStatement(inside);

	inside.close(getTokenReference());
	block.close(getTokenReference());
	self.close(getTokenReference());

	if (block.isReachable() && getMethod().getReturnType() != CStdType.Void) {
	  context.reportTrouble(new CLineError(getTokenReference(), KjcMessages.METHOD_NEED_RETURN, getMethod().getIdent()));
	}
      } else {
	check(context,
	      getMethod().isAbstract()
	      || getMethod().isNative()
	      || context.getClassContext().getCClass().isInterface(),
	      KjcMessages.METHOD_NOBODY_NOABSTRACT, ident);
      }
    } catch (UnpositionedError ce) {
      context.reportTrouble(ce.addPosition(getTokenReference()));
    }
  }

  protected void checkOverriding(CClassContext context, CMethod superMethod) throws PositionedError {
    // check redefinition
    boolean	exactSignature = true;
    CType[]	params1 = superMethod.getParameters();
    CType[]	params2 = ((CSourceMethod)getMethod()).getParameters();
    for (int i = 0; i < params1.length; i++) {
      if (!params1[i].equals(params2[i])) {
	exactSignature = false;
	break;
      }
    }
    if (exactSignature) {
      check(context,
	    returnType.equals(superMethod.getReturnType()),
	    KjcMessages.METHOD_RETURN_DIFFERENT, ident);
      
      checkAccess(context, superMethod);
      
      if (!superMethod.isPrivate()) {
	// JLS 8.4.4
	// A method that overrides or hides another method (JLS 8.4.6), including methods
	// that implement abstract methods defined in interfaces, may not be declared to
	// throw more CHECKED exceptions than the overridden or hidden method.
	
	CClassType[]	exc = superMethod.getThrowables();
	
      loop:
	for (int i = 0; i < exceptions.length; i++) {
	  if (exceptions[i].isCheckedException()) {
	    for (int j = 0; j < exc.length; j++) {
	      if (exceptions[i].isAssignableTo(exc[j])) {
		continue loop;
	      }
	    }
	    check(context, false, KjcMessages.METHOD_THROWS_DIFFERENT, ident, exceptions[i]);
	  }
	}
	
	
	// JLS 8.4.3.3
	// A method can be declared final to prevent subclasses from overriding or hiding it. 
	// It is a compile-time error to attempt to override or hide a final method. 
	check(context, !superMethod.isFinal(), KjcMessages.METHOD_OVERRIDE_FINAL, ident);
      }
    }
  }

  protected void checkAccess(CClassContext context, CMethod superMethod) throws PositionedError {
    check(context,
	  !(getMethod().isPrivate() && !superMethod.isPrivate()),
	  KjcMessages.METHOD_ACCESS_DIFFERENT, ident, null);
    check(context,
	  !(getMethod().isProtected() && !(superMethod.isProtected() || superMethod.isPrivate() || (!superMethod.isPublic()))),
	  KjcMessages.METHOD_ACCESS_DIFFERENT, ident, null);
    check(context,
	  !(!getMethod().isPublic() && superMethod.isPublic()),
	  KjcMessages.METHOD_ACCESS_DIFFERENT, ident, null);
  }

  protected void checkBody(CClassContext context) throws PositionedError {
    check(context,
	  !context.getCClass().isInterface() || (this instanceof JInitializerDeclaration),
	  KjcMessages.METHOD_BODY_IN_INTERFACE, ident);
    
    check(context,
	  !getMethod().isNative() && !getMethod().isAbstract(),
	  KjcMessages.METHOD_BODY_NATIVE_ABSTRACT, ident);
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);
    
    p.visitMethodDeclaration(this,
			     modifiers,
			     returnType,
			     ident,
			     parameters,
			     exceptions,
			     body);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    fail(); // nothing to do here
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  // $$$ MOVE TO BE PRIVATE
  protected int				modifiers;
  protected CType			returnType;
    public     String			ident; //MR made public 
  protected JFormalParameter[]		parameters;
  protected CClassType[]		exceptions;
  protected JBlock			body;
}
