/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JNewObjectExpression.java,v 1.2 2002/03/08 21:54:34 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.*;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a new allocation expression 'new toto(1)'
 */
public class JNewObjectExpression extends JExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	objectType	the type of this object allocator
   * @param	explicitThis	a TypeName expression
   * @param	params		parameters to be passed to constructor
   */
  public JNewObjectExpression(TokenReference where,
			      CClassType objectType,
			      JExpression explicitThis,
			      JExpression[] params) {
    super(where);

    this.type = objectType;
    this.params = params;
    this.explicitThis = explicitThis;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  /**
   * Returns true iff this expression can be used as a statement (JLS 14.7)
   */
  public boolean isStatementExpression() {
    return true;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    CType[] argsType = checkTypes(context);
    
    return checkCall(context, argsType);
  }

  protected JExpression checkCall(CExpressionContext context, CType[] argsType) throws PositionedError {
    context.setDiscardValue(false);

    check(context, !type.getCClass().isAbstract(), KjcMessages.NEW_ABSTRACT, type);
    check(context, !type.getCClass().isInterface(), KjcMessages.NEW_INTERFACE, type);

    try {
      constructor = type.getCClass().lookupMethod(JAV_CONSTRUCTOR, argsType);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }

    if (constructor == null || constructor.getOwner() != type.getCClass()) {
      // May be an inner class
      if (type.getCClass().isInnerClass()) {
	CType[]		argsType2 = new CType[argsType.length + 1];
	System.arraycopy(argsType, 0, argsType2, 0, argsType.length);
	argsType2[argsType.length] = context.getClassContext().getCClass().getType();
	try {
	  constructor = type.getCClass().lookupMethod(JAV_CONSTRUCTOR, argsType2);
	} catch (UnpositionedError e) {
	  throw e.addPosition(getTokenReference());
	}
      }
      if (constructor == null || constructor.getOwner() != type.getCClass()) {
	// do not want a super constructor !
	throw new CMethodNotFoundError(getTokenReference(), null, type.toString(), argsType);
      }
    }

    // check access
    local = context.getClassContext().getCClass();
    check(context, constructor.isAccessible(local), KjcMessages.CONSTRUCTOR_NOACCESS, type);

    if (constructor.getOwner().isInnerClass()) {
      check(context, !constructor.getOwner().hasOuterThis() ||
	    ((!context.getMethodContext().getCMethod().isStatic()) &&
	     (local.descendsFrom(constructor.getOwner().getOwner()) || (local.getOwner() != null && local.getOwner() == constructor.getOwner().getOwner()))),
	    KjcMessages.INNER_INHERITENCE, constructor.getOwner().getType(), local.getType());
    }

    CClassType[]	exceptions = constructor.getThrowables();
    for (int i = 0; i < exceptions.length; i++) {
      context.getBodyContext().addThrowable(new CThrowableInfo(exceptions[i],
							       this,
							       context.getBodyContext().makeSnapshot()));
    }
    
    argsType = constructor.getParameters();

    for (int i = 0; i < params.length; i++) {
      params[i] = params[i].convertType(argsType[i], context);
    }

    return this;
  }

  protected CType[] checkTypes(CExpressionContext context) throws PositionedError {
    try {
      type.checkType(context);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }
    
    CType[] argsType = new CType[params.length];

    for (int i = 0; i < argsType.length; i++) {
      params[i] = params[i].checkExpression(context);
      argsType[i] = params[i].getType();
      assert(argsType[i] != null);
    }

    return argsType;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitNewObjectExpression(this, type, params, null);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    boolean	discardValue = code.discardValue();
    code.setLineNumber(getTokenReference().getLine());

    code.plantInstruction(new ClassRefInstruction(opc_new, type.getCClass().getQualifiedName()));

    if (!discardValue) {
      code.plantInstruction(new NoArgInstruction(opc_dup));
    }

    code.setDiscardValue(false);

    for (int i = 0; i < params.length; i++) {
      params[i].genCode(code);
    }

    if (constructor.getOwner().isInnerClass()) {
      if (!(local.getOwner() != null && local.getOwner() == constructor.getOwner().getOwner())) {
	constructor.getOwner().genSyntheticParams(code);
      }
    }

    if (constructor.getOwner().isInnerClass() && !constructor.getOwner().isStatic() && constructor.getOwner().hasOuterThis()) {
      // SUPER SYNTHETIC PARAMETER
      if (local.getOwner() != null && local.getOwner() == constructor.getOwner().getOwner()) {
	code.plantLoadThis();
	code.plantInstruction(new FieldRefInstruction(opc_getfield,
						      local.getType().getSignature().substring(1, local.getType().getSignature().length() - 1),
						      JAV_OUTER_THIS,
						      local.getOwner().getType().getSignature()));
      }
    }

    code.plantInstruction(new MethodRefInstruction(opc_invokespecial,
						   constructor.getPrefixName(),
						   constructor.getIdent(),
						   constructor.getSignature()));
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected	JExpression		explicitThis;
  protected	JExpression[]		params;
  protected	CClassType		type;
  protected     CClass			local;
  protected	CMethod			constructor;
}
