/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JOrdinalLiteral.java,v 1.2 2002/03/08 21:54:34 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.PushLiteralInstruction;

import at.dms.compiler.NumberParser;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents ordinal literals (byte, short, int, long)
 */
public class JOrdinalLiteral extends JNumberLiteral {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, String image) {
    super(where);
    this.image = image;
  }

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, Number value, CNumericType type) {
    super(where);
    setValue(value, type);
  }

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, long value, CNumericType type) {
    this(where, new Long(value), type);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * This is needed to handle -2147483648 that is allowed agains 2147483648
   * that is not allowed, so we need to know if there is an unary minus before
   * us
   */
  public void setSignToMinus() {
    if (image != null) {
      image = "-" + image;
    }
    if (value != null) {
      value = new Long(-value.longValue());
    }
  }

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    assert(type != null);
    return type;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Number numberValue() {
    assert(value != null);
    return value;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Object getValue() {
    if (getType() == CStdType.Long) {
      return value;
    } else {
      return new Integer(value.intValue());
    }
  }

  /**
   * @param	newValue		the new value
   */
  private void setValue(Number newValue, CNumericType newType) {
    long	val = newValue.longValue();

    if (val == newValue.byteValue()) {
      type = CStdType.Byte;
      value = new Byte(newValue.byteValue());
    } else if (val == newValue.shortValue()) {
      type = CStdType.Short;
      value = new Short(newValue.shortValue());
    } else if (val <= Character.MAX_VALUE && val >= Character.MIN_VALUE) {
      type = CStdType.Char;
      value = new Integer(newValue.intValue());
    } else if (val == newValue.intValue()) {
      type = CStdType.Integer;
      value = new Integer(newValue.intValue());
    } else {
      type = CStdType.Long;
      value = newValue;
    }

    type = newType;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    if (value == null) {
      try {
	if ((image.indexOf("L") > 0) || (image.indexOf("l") > 0)) {
	  setValue(new Long(NumberParser.decodeLong(image.substring(0, image.length() - 1))), CStdType.Long);
	} else {
	  setValue(new Integer(NumberParser.decodeInt(image)), CStdType.Integer);
	}
	image = null;
      } catch (NumberFormatException ex) {
	check(context, false, KjcMessages.LITER_INT_BAD, image);
      }
    }

    return this;
  }

  /**
   * convertType
   * changes the type of this expression to an other
   * @param  dest the destination type
   */
  public JExpression convertType(CType dest, CExpressionContext context) throws PositionedError {
    assert(type != null);
    if (type.equals(dest)) {
      return this;
    } else if (dest == CStdType.Byte) {
      return new JOrdinalLiteral(getTokenReference(), new Byte((byte)value.longValue()), CStdType.Byte);
    } else if (dest == CStdType.Short) {
      return new JOrdinalLiteral(getTokenReference(), new Short((short)value.longValue()), CStdType.Short);
    } else if (dest == CStdType.Char) {
      return new JOrdinalLiteral(getTokenReference(), new Integer((int)value.longValue()), CStdType.Char);
    } else if (dest == CStdType.Integer) {
      return new JOrdinalLiteral(getTokenReference(), new Integer((int)value.longValue()), CStdType.Integer);
    } else if (dest == CStdType.Long) {
      return new JOrdinalLiteral(getTokenReference(), new Long(value.longValue()), CStdType.Long);
    } else if (dest == CStdType.Float) {
      return new JRealLiteral(getTokenReference(), new Float((float)value.longValue()), CStdType.Float);
    } else if (dest == CStdType.Double) {
      return new JRealLiteral(getTokenReference(), new Double((double)value.longValue()), CStdType.Double);
    } else {
      fail();
      return null;
    }
  }

  /**
   * Can this expression be converted to the specified type by assignment conversion (JLS 5.2) ?
   * @param	dest		the destination type
   * @return	true iff the casting conversion is valid
   */
  public boolean isAssignableTo(CType dest) {
    if (type != CStdType.Integer) {
      return type.isAssignableTo(dest);
    } else {
      if (dest == CStdType.Byte) {
	long	val = value.longValue();

	return (byte)val == val;
      } else if (dest == CStdType.Short) {
	long	val = value.longValue();

	return (short)val == val;
      } else if (dest == CStdType.Char) {
	long	val = value.longValue();

	return (char)val == val;
      } else {
	return type.isAssignableTo(dest);
      }
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    assert(value != null);

    if (type == CStdType.Byte) {
      p.visitByteLiteral(value.byteValue());
    } else if (type == CStdType.Integer) {
      p.visitIntLiteral(value.intValue());
    } else if (type == CStdType.Long) {
      p.visitLongLiteral(value.longValue());
    } else if (type == CStdType.Short) {
      p.visitShortLiteral(value.shortValue());
    } else if (type == CStdType.Char) {
	;
    } else
	{
	    System.out.println( type );
      fail();
	}
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (type == CStdType.Long) {
      code.plantInstruction(new PushLiteralInstruction(numberValue().longValue()));
    } else {
      code.plantInstruction(new PushLiteralInstruction(numberValue().intValue()));
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected String		image;
  private Number		value;
  private CType			type;
}
