/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JPostfixExpression.java,v 1.2 2002/03/08 21:54:35 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.IincInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents postfix increment and decrement expressions.
 * 15.13.2 Postfix Increment Operator ++
 * 15.13.3 Postfix Decrement Operator --
 */
public class JPostfixExpression extends JExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	oper		the operator
   * @param	expr		the operand
   */
  public JPostfixExpression(TokenReference where, int oper, JExpression expr) {
    super(where);
    this.oper = oper;
    this.expr = expr;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  /**
   * Returns true iff this expression can be used as a statement (JLS 14.7)
   */
  public boolean isStatementExpression() {
    return true;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    expr = expr.checkExpression(context, true, context.discardValue());
    check(context, expr.getType().isNumeric(), KjcMessages.POSTFIX_BADTYPE, expr.getType());
    check(context,
	  (expr instanceof CInitializable) && ((CInitializable)expr).isLValue(context),
	  KjcMessages.POSTFIX_NOTLVALUE);
    check(context,
	  ((CInitializable)expr).isInitialized(context),
	  KjcMessages.POSTFIX_NOTINITIALIZED);

    type = expr.getType();

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitPostfixExpression(this, oper, expr);
  }


  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());
    boolean		discardValue = code.discardValue();
    int			val = oper == OPE_POSTINC ? 1 : -1;

    if ((expr.getType() == CStdType.Integer) &&
	(expr instanceof JLocalVariableExpression)) {
      expr.genCode(code);
      code.plantInstruction(new IincInstruction(((JLocalVariableExpression)expr).getPosition(),
						val));
    } else {
      if (discardValue) {
	code.setDiscardValue(false);
	((CInitializable)expr).genStartStoreCode(code);
	expr.genCode(code);
      }  else {
	((CInitializable)expr).genStartStoreCode(code);
	expr.genCode(code);
	if (getType().getSize() == 2) {
	  if ((expr instanceof JLocalVariableExpression) ||
	      ((expr instanceof JClassFieldExpression) &&
	       ((JClassFieldExpression)expr).getField().isStatic())) {
	    code.plantInstruction(new NoArgInstruction(opc_dup2));
	  } else if (expr instanceof JArrayAccessExpression) {
	    code.plantInstruction(new NoArgInstruction(opc_dup2_x2));
	  } else {
	    code.plantInstruction(new NoArgInstruction(opc_dup2_x1));
	  }
	} else if (getType().getSize() == 1) {
	  if ((expr instanceof JLocalVariableExpression) ||
	      ((expr instanceof JClassFieldExpression) &&
	       ((JClassFieldExpression)expr).getField().isStatic())) {
	    code.plantInstruction(new NoArgInstruction(opc_dup));
	  } else if (expr instanceof JArrayAccessExpression) {
	    code.plantInstruction(new NoArgInstruction(opc_dup_x2));
	  } else {
	    code.plantInstruction(new NoArgInstruction(opc_dup_x1));
	  }
	}
      }

      switch (expr.getType().getTypeID()) {
      case TID_FLOAT:
	code.plantInstruction(new PushLiteralInstruction((float)val));
	code.plantInstruction(new NoArgInstruction(opc_fadd));
	break;
      case TID_LONG:
	code.plantInstruction(new PushLiteralInstruction((long)val));
	code.plantInstruction(new NoArgInstruction(opc_ladd));
	break;
      case TID_DOUBLE:
	code.plantInstruction(new PushLiteralInstruction((double)val));
	code.plantInstruction(new NoArgInstruction(opc_dadd));
	break;

      case TID_BYTE:
	code.plantInstruction(new PushLiteralInstruction(val));
	code.plantInstruction(new NoArgInstruction(opc_iadd));
	code.plantInstruction(new NoArgInstruction(opc_i2b));
	break;
      case TID_CHAR:
	code.plantInstruction(new PushLiteralInstruction(val));
	code.plantInstruction(new NoArgInstruction(opc_iadd));
	code.plantInstruction(new NoArgInstruction(opc_i2c));
	break;
      case TID_SHORT:
	code.plantInstruction(new PushLiteralInstruction(val));
	code.plantInstruction(new NoArgInstruction(opc_iadd));
	code.plantInstruction(new NoArgInstruction(opc_i2s));
	break;
      case TID_INT:
	code.plantInstruction(new PushLiteralInstruction(val));
	code.plantInstruction(new NoArgInstruction(opc_iadd));
	break;
      }

      ((CInitializable)expr).genEndStoreCode(code, true);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected int			oper;
  protected JExpression		expr;
  protected CType		type;
}

