/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JRealLiteral.java,v 1.2 2002/03/08 21:54:36 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Root class for all expressions
 */
public class JRealLiteral extends JNumberLiteral {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JRealLiteral(TokenReference where, String image) {
    super(where);
    this.image = image;
  }

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JRealLiteral(TokenReference where, Number value, CType type) {
    super(where);
    assert(type == CStdType.Double || type == CStdType.Float);
    setValue(value);
    this.type = type;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Number numberValue() {
    assert(value != null);
    return value;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Object getValue() {
    return numberValue();
  }

  /**
   * @param	newValue		the new value
   */
  private void setValue(Number newValue) {
    type = CStdType.Double;
    value = newValue;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    Double	doubleObject = null;

    if (value != null) {
      return this; // already checked
    }

    try {
      doubleObject = Double.valueOf(image);
    } catch (NumberFormatException ex) {
      check(context, false, KjcMessages.LITER_DOUBLE_BAD, image);
    }
    setValue(doubleObject);

    if (image.endsWith("f") || image.endsWith("F")) {
      type = CStdType.Float;
    }

    return this;
  }

  /**
   * convertType
   * changes the type of this expression to an other
   * @param  dest the destination type
   */
  public JExpression convertType(CType dest, CExpressionContext context) throws PositionedError {
    if (! type.equals(dest)) {
      if (dest == CStdType.Float) {
	assert(type == CStdType.Double);
	value = new Float((float)value.doubleValue());
      } else if (dest == CStdType.Double) {
	assert(type == CStdType.Float);
	value = new Double((double)value.floatValue());
      } else {
	fail();
      }
      type = dest;
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    assert(value != null);
    if (type == CStdType.Double) {
      p.visitDoubleLiteral(value.doubleValue());
    } else if (type == CStdType.Float) {
      p.visitFloatLiteral(value.floatValue());
    } else {
      fail();
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (type == CStdType.Float) {
      code.plantInstruction(new PushLiteralInstruction(value.floatValue()));
    } else {
      code.plantInstruction(new PushLiteralInstruction(value.doubleValue()));
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private String		image;
  private Number		value;
  private CType			type;
}
