/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JReturnStatement.java,v 1.2 2002/03/08 21:54:36 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.15: Return Statement
 *
 * A return statement returns control to the invoker of a method or constructor.
 */
public class JReturnStatement extends JStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

 /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	expr		the expression to return.
   */
  public JReturnStatement(TokenReference where, JExpression expr, JavaStyleComment[] comments) {
    super(where, comments);
    this.expr = expr;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------


  /**
   * Returns the type of this return statement
   */
  public CType getType() {
    return expr != null ? expr.getType() : CStdType.Void;
  }


  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    CType	returnType = context.getMethodContext().getCMethod().getReturnType();

    if (expr != null) {
      check(context, returnType != CStdType.Void, KjcMessages.RETURN_NONEMPTY_VOID);

      CExpressionContext	expressionContext = context.createExpressionContext();

      expr = expr.checkExpression(expressionContext);
      check(context,
	    expr.isAssignableTo(returnType),
	    KjcMessages.RETURN_BADTYPE, expr.getType(), returnType);
      expr = expr.convertType(returnType, expressionContext);
    } else {
      check(context, returnType == CStdType.Void, KjcMessages.RETURN_EMPTY_NONVOID);
    }

    context.addReturn();
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);
    p.visitReturnStatement(this, expr);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (expr != null) {
      expr.genCode(code);

      code.plantReturn(this);
      code.plantInstruction(new NoArgInstruction(expr.getType().getReturnOpcode()));
    } else {
      code.plantReturn(this);
      code.plantInstruction(new NoArgInstruction(opc_return));
    }
  }

  /**
   * Load the value from a local var (after finally)
   */
  public void load(CodeSequence code, JLocalVariable var) {
    code.plantLocalVar(expr.getType().getLoadOpcode(), var);
  }

  /**
   * Load the value from a local var (after finally)
   */
  public void store(CodeSequence code, JLocalVariable var) {
    code.plantLocalVar(expr.getType().getStoreOpcode(), var);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected JExpression		expr;
}
