/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSwitchStatement.java,v 1.2 2002/03/08 21:54:37 mrobilla Exp $
 */

package at.dms.kjc;

import java.util.Vector;

import at.dms.classfile.SwitchInstruction;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;
import at.dms.util.VectorCache;

/**
 * JLS 14.9: Switch Statement
 */
public class JSwitchStatement extends JStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	expr		the expr part
   * @param	groups		the different part of body.
   */
  public JSwitchStatement(TokenReference where,
			  JExpression expr,
			  JSwitchGroup[] groups,
			  JavaStyleComment[] comments)
  {
    super(where, comments);

    this.expr = expr;
    this.groups = groups;
    this.hasBreak = false;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    expr = expr.checkExpression(context.createExpressionContext());
    check(context, expr.getType().implicitlyCastTo(CStdType.Integer), KjcMessages.SWITCH_BADEXPR);

    CSwitchBodyContext	self = new CSwitchBodyContext(this, context, expr.getType());
    CBodyContext	inside = new CBodyContext(self);
    CBodyContext[]	groupContexts = new CBodyContext[groups.length];

    for (int i = 0; i < groups.length; i++) {
      if (i == 0 || !groupContexts[i - 1].isReachable()) {
	groupContexts[i] = inside.cloneContext();
      } else {
	groupContexts[i] = groupContexts[i - 1];
      }
      groups[i].checkBody(self, groupContexts[i]);
    }

    boolean	isFirst = self.defaultExists();

    for (int i = 0; i < groupContexts.length; i++) {
      if (isFirst && (groupContexts[i].isReachable() || groupContexts[i].isBreaked())) {
	inside.adopt(groupContexts[i]);
	isFirst = false;
      } else if (groupContexts[i].isBreaked() || 
		 i == groupContexts.length - 1 && groupContexts[i].isReachable()) {
	inside.merge(groupContexts[i]);
      }

      if (groupContexts[i].isReachable() && i != groupContexts.length - 1) {
	context.reportTrouble(new CWarning(getTokenReference(),
					   KjcMessages.CASE_FALL_THROUGH));
      }
    }
    inside.close(getTokenReference());
    self.setReachable(groupContexts[groupContexts.length - 1].isReachable() ||
		      hasBreak ||
		      !self.defaultExists());

    self.close(getTokenReference());
    
    if (!self.defaultExists()) {
      context.reportTrouble(new CWarning(getTokenReference(),
					 KjcMessages.SWITCH_NO_DEFAULT));
    }
  }

  // ----------------------------------------------------------------------
  // BREAK/CONTINUE HANDLING
  // ----------------------------------------------------------------------

  /**
   * Returns true iff this statement is a possible target for a break statement.
   */
  public boolean acceptsBreak() {
    return true;
  }

  /**
   * Returns true iff this statement is a possible target for a continue statement.
   */
  public boolean acceptsContinue() {
    return false;
  }

  /**
   * Prevent statement that there is at least one break
   */
  public void addBreak(CBodyContext context) {
    hasBreak = true;
  }

  /**
   * Return the end of this block (for break statement)
   */
  public CodeLabel getBreakLabel() {
    return endLabel;
  }

  /**
   * Return the beginning of this block (for continue statement)
   */
  public CodeLabel getContinueLabel() {
    fail();
    return null;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);
    p.visitSwitchStatement(this, expr, groups);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    CodeLabel		defaultLabel = new CodeLabel();

    Vector		matches = VectorCache.request(VectorCache.BIG);
    Vector		targets = VectorCache.request(VectorCache.BIG);

    for (int i = 0; i < groups.length; i++) {
      groups[i].collectLabels(defaultLabel, matches, targets);
    }

    expr.genCode(code);

    code.pushContext(this);
    code.plantInstruction(new SwitchInstruction(defaultLabel, matches, targets));
    for (int i = 0; i < groups.length; i++) {
      groups[i].genCode(code);
    }

    if (!defaultLabel.hasAddress()) {
      code.plantLabel(defaultLabel);
    }

    code.plantLabel(endLabel);
    code.popContext(this);

    VectorCache.release(matches);
    VectorCache.release(targets);

    endLabel = null;
  }


  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private JExpression		expr;
  private JSwitchGroup[]	groups;
  private boolean		hasBreak;
  private CodeLabel		endLabel = new CodeLabel();
}
