/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTryFinallyStatement.java,v 1.2 2002/03/08 21:54:38 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.HandlerInfo;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.18: Try Statement
 *
 * A try statement executes a block.
 * If a value is thrown and the try statement has one or more catch
 * clauses that can catch it, then control will be transferred to the
 * first such catch clause.
 * If the try statement has a finally clause, then another block of code
 * is executed, no matter whether the try block completes normally or abruptly,
 * and no matter whether a catch clause is first given control.
 *
 * In this implementation, the Try Statement is split into a Try-Catch Statement
 * and a Try-Finally Statement. A Try Statement where both catch and finally
 * clauses are present is rewritten as a Try-Catch Statement enclosed in a
 * Try-Finally Statement.
 */
public class JTryFinallyStatement extends JStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where			the line of this node in the source code
   * @param	tryClause		the body
   * @param	finallyClause		the finally clause
   */
  public JTryFinallyStatement(TokenReference where,
			      JBlock tryClause,
			      JBlock finallyClause,
			      JavaStyleComment[] comments)
  {
    super(where, comments);

    this.tryClause = tryClause;
    this.finallyClause = finallyClause;
    this.finallyLabel = new CodeLabel();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    localVar = new JGeneratedLocalVariable(null, 
					   0,
					   CStdType.Integer,
					   "try" + toString() /* unique ID */,
					   null);
    try {
      context.getBlockContext().addVariable(localVar);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }
    finallyVar = new JGeneratedLocalVariable(null,
					     0,
					     CStdType.Integer,
					     "try_finally_" + toString() /* unique ID */,
					     null);
    try {
      context.getBlockContext().addVariable(finallyVar);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }
    returnVar = new JGeneratedLocalVariable(null,
					    0,
					    CStdType.Integer,
					    "try_finally_ret" + toString() /* unique ID */,
					    null);
    try {
      context.getBlockContext().addVariable(returnVar);
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }

    CTryContext		tryContext = new CTryContext(context);
    CBodyContext	inside = new CBodyContext(tryContext);

    CTryContext		finallyContext = new CTryContext(context);
    CBodyContext	finallyInside = new CBodyContext(finallyContext);

    tryClause.checkStatement(inside);
    inside.close(getTokenReference());

    // finally always reachable
    finallyClause.checkBody(finallyInside);
    finallyInside.close(getTokenReference());

    if (!finallyContext.isReachable()) {
      tryContext.setReachable(finallyContext.isReachable());
      tryContext.setThrowables(finallyContext.getThrowables());
    } else {
      tryContext.addThrowables(finallyContext.getThrowables());
    }

    tryContext.close(getTokenReference());
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * @return the finally label or null
   */
  public CodeLabel getFinallyLabel() {
    return finallyLabel;
  }

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitTryFinallyStatement(this, tryClause, finallyClause);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    CodeLabel		nextLabel	= new CodeLabel();

    int		startPC = code.getPC();
    code.pushContext(this);
    tryClause.genCode(code);			//		TRY CODE
    int		endPC = code.getPC();
    code.plantLabelRef(opc_jsr, finallyLabel);	//		JSR finally
    code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next

    // HANDLER FOR OTHER EXCEPTIONS
    int	finallyPC = code.getPC();
    code.plantLocalVar(opc_astore, localVar);
    code.plantLabelRef(opc_jsr, finallyLabel);	//		JSR finally
    code.plantLocalVar(opc_aload, localVar);
    code.plantInstruction(new NoArgInstruction(opc_athrow));

    // FINALLY CODE
    code.plantLabel(finallyLabel);		//	finally:...
    code.plantLocalVar(opc_astore, finallyVar);
    finallyClause.genCode(code);			//		FINALLY CLAUSE

    code.plantLocalVar(opc_ret, finallyVar);
    code.addExceptionHandler(new HandlerInfo(code.getInstructionAt(startPC),
					     code.getInstructionAt(endPC - 1 == -1 ? 0 : endPC - 1),
					     code.getInstructionAt(finallyPC),
					     null));
					     //CStdType.Throwable.getCClass().getQualifiedName()));
    code.plantLabel(nextLabel);			//	next:	...
    code.popContext(this);

    finallyLabel = null;
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genFinallyCall(CodeSequence code, JReturnStatement ret) {
    if (ret != null && ret.getType().getSize() > 0) {
      ret.store(code, returnVar);
      code.plantLabelRef(Constants.opc_jsr, finallyLabel);
      ret.load(code, returnVar);
    } else {
      code.plantLabelRef(Constants.opc_jsr, finallyLabel);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private CodeLabel		finallyLabel;
  private JBlock		tryClause;
  private JBlock		finallyClause;

  private JLocalVariable	localVar;
  private JLocalVariable	finallyVar;
  private JLocalVariable	returnVar;
}
