/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package at.dms.kjc;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import ca.ubc.cs.jex.JexFile;
import ca.ubc.cs.jex.Workspace;
import ca.ubc.cs.jex.JexFileCollection;
import ca.ubc.cs.jex.ExceptionBlock;
import ca.ubc.cs.jex.ExceptionThrow;
import ca.ubc.cs.jex.ExceptionGet;
import ca.ubc.cs.jex.TryBlock;
import ca.ubc.cs.jex.JexMethod;
import ca.ubc.cs.jex.AnonymousJexFile;
import ca.ubc.cs.jex.file.ParseException;
import ca.ubc.cs.jex.file.TokenMgrError;
import ca.ubc.cs.jex.Log;
import java.util.Stack;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.lang.reflect.Method;
import ca.ubc.cs.jex.util.MethodRelationshipsException;

/**
 * Implementation of Visitor Design Pattern for KJC.
 */
public class JexVisitor implements KjcVisitor
{
    private JexFile           aCurrentJexFile;
    private Stack             aCurrentExceptionBlock;
    private Workspace         aWorkspace;
    private JexFileCollection aJexFiles;
    private Log               aLog;

    public JexVisitor( Workspace pWorkspace )
    {
	aWorkspace = pWorkspace;
	aCurrentExceptionBlock = new Stack();
	aLog = new Log();
    }

    public void clear()
    {
	aCurrentJexFile = null;
	aCurrentExceptionBlock = new Stack();
	aJexFiles = null;
	aLog = new Log();
    }

    public JexFileCollection getJexFiles()
    { return aJexFiles; }

    public Log getLog()
    { return aLog; }

    public void iterateJexFiles()
    {
	Iterator lIt = aJexFiles.iterator();
	while( lIt.hasNext() )
	{
	    aCurrentJexFile = ((JexFile)lIt.next()).copy();
	    JClassDeclaration lDecl = aJexFiles.getClassDecl( aCurrentJexFile.getQualifiedName().replace('.','/') );
	    lDecl.accept( this );
	    aJexFiles.addJexFile( aCurrentJexFile.getQualifiedName().replace('.','/'), aCurrentJexFile );
	}
    }

    public void iterateJexClass( JClassDeclaration self,
				 JMethodDeclaration[] methods )
    {
	aCurrentExceptionBlock.push( new ExceptionBlock() );

	JFieldDeclaration[] lFields = self.getFields();
	if( lFields != null )
	    for( int i = 0; i < lFields.length; i++ )
		lFields[i].accept( this );
	
	JInitializerDeclaration lInstanceInit = self.instanceInit;
	if( lInstanceInit != null )
	{
	    lInstanceInit.body.accept( this );
	}

	// Pop the initializer block
	ExceptionBlock lNewBlock = (ExceptionBlock) aCurrentExceptionBlock.pop();
	aCurrentJexFile.setInitializerBlock( lNewBlock );

	// Process the constructors & methods
	
	JConstructorDeclaration lDefault = self.getDefaultConstructor();
	if( lDefault != null )
	{
	    lDefault.accept( this );
	}
	
	if( methods != null )
	{
	    for( int i = 0; i < methods.length; i++ )
	    {
		JexMethod lMethod;
		if( methods[i] instanceof JConstructorDeclaration )
		    lMethod = aCurrentJexFile.getJexMethod( makeConstructorSignature( methods[i].parameters ));
		else
		    lMethod = aCurrentJexFile.getJexMethod( makeMethodSignature( methods[i].ident, methods[i].parameters ));
		
		if( lMethod != null )
		{
		    methods[i].accept( this );
		}
		else
		{
		    aLog.addError( aCurrentJexFile.getQualifiedName(), "Cannot find internal method " +
				   aCurrentJexFile.getQualifiedName() + "." + 
				   makeMethodSignature( methods[i].ident, methods[i].parameters ));
		}
	    }
	}
    }

    private void addEnvironmentException( String pException )
    {
	try
        {
	    ((ExceptionBlock)aCurrentExceptionBlock.peek()).addElement( new ExceptionGet( Class.forName( pException ) ));
	}
	catch( ClassNotFoundException e )
        { 
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Invalid class " + pException );
	}
    }

    private void addCallException( Class pException, String pOrigin )
    {
	((ExceptionBlock)aCurrentExceptionBlock.peek()).addElement( new ExceptionGet( pException, pOrigin ));
    }

    /** pClass name is using the slash format */
    /**
    private void addAllInternalNonVirtualCallExceptions( String pClassName, String pSignature )
    {
	JexFile lJexFile = (JexFile)aJexFiles.getJexFile( pClassName );
	
	if( lJexFile == null )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Can't find internal class " + pClassName );
	    return;
	}

	try
	{
	    Set lExceptions = lJexFile.getExceptions( pSignature );
	    Iterator lList = lExceptions.iterator();
	    while( lList.hasNext() )
	    {
		Class lExc = (Class)lList.next();
		addCallException( lExc, pClassName.replace('/','.') + "." + pSignature );
	    }
	}
	catch( NoSuchMethodException e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Can't find internal method " + 
			   pClassName + "." + pSignature );
	}
    }*/

    private void addExternalNonVirtualCallExceptions( String pClassName, String pSignature )
    {
	// We just don't do local anonymous classes
	if( aJexFiles.contains( pClassName ) )
	    if( aJexFiles.getJexFile( pClassName ).isAnonymous() )
		return;

	try
	{
	    Set lExceptions = aWorkspace.getExceptionsForNonVirtualCall( pClassName, pSignature );
	    Iterator lList = lExceptions.iterator();
	    while( lList.hasNext() )
	    {
		Class lExc = (Class)lList.next();
		addCallException( lExc, pClassName.replace('/','.') + "." + pSignature );
	    }
	}
	catch( FileNotFoundException e )
	{
	    aLog.addJexFile( pClassName );
	}
	catch( NoSuchMethodException e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Method " + pSignature + 
			   " not found in Jex file for " + pClassName );
	}
	catch( ParseException e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Jex file for class " + pClassName + 
			   " has invalid syntax: " + e.getMessage() );
	}
	catch( ClassNotFoundException e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Invalid class " + e.getMessage() );
	}
	catch( TokenMgrError e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Jex file for class " + pClassName + 
			   " has invalid syntax: " + e.getMessage() );
	}
    }
    
    /** pClassName is using the dot format */
    private void addVirtualCallExceptions( String pClassName, String pSignature )
    {
	try
	{
	    try
	    {
		Vector lImpls = aWorkspace.getImplementations( pClassName, pSignature );
		for( int i = 0; i < lImpls.size(); i++ )
		{
		    Method lMethod = (Method) lImpls.elementAt(i);
		    String lClassName = lMethod.getDeclaringClass().getName().replace('.','/');
		    //if( aJexFiles.contains( lClassName ))
		    //{
		    //	addAllInternalNonVirtualCallExceptions( lClassName.replace('.','/'), pSignature );
		    //}
		    //else
		    //{
		    addExternalNonVirtualCallExceptions( lClassName, pSignature );
		    //}
		}
	    }
	    catch( MethodRelationshipsException e )
	    {
		aLog.addMethod( pClassName + "." + pSignature );

		//if( aJexFiles.contains( pClassName ))
		//    addAllInternalNonVirtualCallExceptions( pClassName.replace('.','/'), pSignature );
		//else
		    addExternalNonVirtualCallExceptions( pClassName, pSignature );
	    }
	    addAnonymousVirtualCallExceptions( pClassName, pSignature );
	}
	catch( ClassNotFoundException e )
	{
	    //System.out.println( "In addVirtualCallExceptions: " + pClassName + "." + pSignature );
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Invalid class " + e.getMessage() );
	}
	catch( NoSuchMethodException e )
	{
	    //System.out.println( "in addVirtualCallExceptions: " + pClassName + "." + pSignature );
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Trying to load unexisting method " + e.getMessage() );
	}
	
    }

    private void addAnonymousVirtualCallExceptions( String pClassName, String pSignature )
    {
	try
	{
	    Set lExceptions = aWorkspace.getExceptionsFromAnonymousClasses( pClassName, pSignature );
	    Iterator lList = lExceptions.iterator();
	    while( lList.hasNext() )
	    {
		Class lExc = (Class)lList.next();
		addCallException( lExc, "anonymous" + "." + pSignature );
	    }
	}
	//catch( NoSuchMethodException e )
	//{
	//   System.out.println( "in addAnonymousVirtualCallExceptions: " + pClassName + "." + pSignature );
	//   aLog.addError( aCurrentJexFile.getQualifiedName(), "Trying to load unexisting method " + e.getMessage() );
	//}
	catch( IOException e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "IO Exception while loading anonymous classe for " 
			   + pClassName +"." + pSignature );
	}
	catch( ClassNotFoundException e )
	{
	    //System.out.println( "In addAnonymousVirtualCallExceptions: " + pClassName +"." + pSignature );
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Invalid class " + e.getMessage() );
	}
	catch( ParseException e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Jex file for anonymous subclass of " + pClassName + 
			   " has invalid syntax: " + e.getMessage() );
	}
	catch( TokenMgrError e )
	{
	    aLog.addError( aCurrentJexFile.getQualifiedName(), "Jex file for anonymous subclass of " + pClassName + 
			   " has invalid syntax: " + e.getMessage() );
	}
    }

    public static String toJexType( CType pType )
    {
	if( pType.isArrayType() )
	    return pType.getSignature().replace('/','.');
	else if( pType.isPrimitive() )
	    return pType.toString();
	else
	    return pType.getCClass().getQualifiedName().replace('/','.');
    }

    public static String makeMethodSignature( String pIdent, JFormalParameter[] pParams )
    {
	String lReturn = pIdent + "(";

	for( int i = 0; i < pParams.length -1; i++ )
	    lReturn += toJexType( pParams[i].getType() ) + ",";
	
	if( pParams.length > 0 )
	    lReturn += toJexType( pParams[pParams.length - 1].getType() );

	lReturn += ")";
	return lReturn;
    }

    public static String makeConstructorSignature( JFormalParameter[] pParams )
    {
	return makeMethodSignature( "<init>", pParams );
    }

    public static String makeMethodSignature( String pIdent, CType[] pParams )
    {
	String lReturn = pIdent + "(";

	for( int i = 0; i < pParams.length -1; i++ )
	    lReturn += toJexType( pParams[i] ) + ",";
	
	if( pParams.length > 0 )
	    lReturn += toJexType( pParams[pParams.length - 1] );

	lReturn += ")";
	return lReturn;
    }

    public static String makeConstructorSignature( CType[] pParams )
    {
	return makeMethodSignature( "<init>", pParams );
    }

    // ----------------------------------------------------------------------
    // COMPILATION UNIT
    // ----------------------------------------------------------------------
    
    /**
     * visits a compilation unit
     */
    public void visitCompilationUnit(JCompilationUnit self,
				     JPackageName packageName,
				     JPackageImport[] importedPackages,
				     JClassImport[] importedClasses,
				     JTypeDeclaration[] typeDeclarations)
    {
	// Collect all the class names for this compilation unit
	TypeDeclarationCollectorVisitor lCollector = new TypeDeclarationCollectorVisitor();
	self.accept( lCollector );
	aJexFiles = lCollector.getJexFileCollection();

	//System.out.println( "Analyzing " + self.getTokenReference().getFile() ); //DB
	// Do first pass

	int lLastCheckSum = 0;

	iterateJexFiles();
	  
	/**
	if( aJexFiles.size() > 1 )
	{
	    int lCheckSum = aJexFiles.getCheckSum();

	    while( (lCheckSum - lLastCheckSum) != 0 )
	    {
		lLastCheckSum = lCheckSum;
		iterateJexFiles();
		lCheckSum = aJexFiles.getCheckSum();
	    }
	    }*/
    }

  // ----------------------------------------------------------------------
  // TYPE DECLARATION
  // ----------------------------------------------------------------------

    /**
     * visits a class declaration
     */
    public void visitClassDeclaration(JClassDeclaration self,
				      int modifiers,
				      String ident,
				      String superName,
				      CClassType[] interfaces,
				      JPhylum[] body,
				      JMethodDeclaration[] methods,
				      JTypeDeclaration[] decls)
    {
	
	int lLastCheckSum = 0;
	//System.out.println( "   " + self.getCClass().getQualifiedName() + "(1)");
	iterateJexClass( self, methods );
	//int lCheckSum = aCurrentJexFile.getCheckSum();
	//int lLoop = 2;

	//while( (lCheckSum - lLastCheckSum) != 0 )
	//{
	//   lLastCheckSum = lCheckSum;
	//   iterateJexClass( self, methods );
	//   System.out.println( "   " + self.getCClass().getQualifiedName() + "(" + lLoop + ")" );
	//   lLoop++;
	//   lCheckSum = aCurrentJexFile.getCheckSum();
	//}
	
	// Do not process the inners, they will be done by iteration
    }
  
    // ----------------------------------------------------------------------
    // METHODS AND FIELDS
    // ----------------------------------------------------------------------
    
    /**
     * visits a field declaration
     */
    public void visitFieldDeclaration(JFieldDeclaration self,
				      int modifiers,
				      CType type,
				      String ident,
				      JExpression expr)
    {
	// Add the initialization to the current Jexfile's constructor prefixes.
	if( expr != null )
	    expr.accept( this );
    }

  /**
   * visits a method declaration
   */
  public void visitMethodDeclaration(JMethodDeclaration self,
			      int modifiers,
			      CType returnType,
			      String ident,
			      JFormalParameter[] parameters,
			      CClassType[] exceptions,
			      JBlock body)
  {

      JexMethod lJexMethod = aCurrentJexFile.getJexMethod( makeMethodSignature( ident, parameters) );
      if( lJexMethod == null )
      {
	  aLog.addError( aCurrentJexFile.getQualifiedName(), "Could not find internal method " +
			 aCurrentJexFile.getQualifiedName() + "." + makeMethodSignature( ident, parameters ));
	  return;
      }

      aCurrentExceptionBlock.push( new ExceptionBlock() );

      if( body != null )
	  body.accept( this );

      ExceptionBlock lNewBlock = (ExceptionBlock)aCurrentExceptionBlock.pop();
      lJexMethod.setExceptionBlock( lNewBlock );
  }

  /**
   * visits a method declaration
   */
  public void visitConstructorDeclaration(JConstructorDeclaration self,
				   int modifiers,
				   String ident,
				   JFormalParameter[] parameters,
				   CClassType[] exceptions,
				   JBlock body)
      {
	  
	  JexMethod lJexMethod = aCurrentJexFile.getJexMethod( makeConstructorSignature( parameters ) );
	  if( lJexMethod == null )
	  {
	      aLog.addError( aCurrentJexFile.getQualifiedName(), "Could not find internal method " +
			     aCurrentJexFile.getQualifiedName() + "." + makeConstructorSignature( parameters ));
	      return;
	  }

	  aCurrentExceptionBlock.push( new ExceptionBlock() );

	  if( body != null )
	  {
	      // create the exceptions for the super call
	      JExplicitConstructorInvocation lExplicit = ((JConstructorBlock)body).explicitSuper;
	      lExplicit.accept( this );
	      body.accept( this );
	  }

	  ExceptionBlock lBlock = (ExceptionBlock)aCurrentExceptionBlock.pop();
	  lJexMethod.setExceptionBlock( lBlock );
      }

  // ----------------------------------------------------------------------
  // STATEMENTS
  // ----------------------------------------------------------------------

  /**
    * visits a while statement
    * Adds the exceptions of the expression and statement to the current list.
    */
    public void visitWhileStatement(JWhileStatement self,
				    JExpression cond,
				    JStatement body)
    {

	if( cond != null ) cond.accept( this );
	if( body != null ) body.accept( this );
    }

    /**
      * visits a variable declaration statement
      * adds any exceptions potentially raised in initialization
      */
    public void visitVariableDeclarationStatement(JVariableDeclarationStatement self,
						  JVariableDefinition[] vars)
    {

	if( vars != null )
	    for( int i = 0; i < vars.length; i++ )
		vars[i].accept( this );
    }

    /**
     * visits a variable declaration statement
     */
    public void visitVariableDefinition(JVariableDefinition self,
					int modifiers,
					CType type,
					String ident,
					JExpression expr)
    {
	if( expr != null ) expr.accept( this );
    }
  /**
   * visits a try-catch statement
   */
    public void visitTryCatchStatement(JTryCatchStatement self,
				       JBlock tryClause,
				       JCatchClause[] catchClauses)
    {
	TryBlock lBlock = new TryBlock();
	((ExceptionBlock)aCurrentExceptionBlock.peek()).addElement( lBlock );
	aCurrentExceptionBlock.push( lBlock.getTryBlock() );
	if( tryClause != null ) tryClause.accept( this );
	aCurrentExceptionBlock.pop();

	if( catchClauses != null )
	    for( int i = 0; i < catchClauses.length; i++ )
	    {
		try
		{
		    ExceptionBlock lCatchBlock = lBlock.addCatchBlock( Class.forName( catchClauses[i].getType().toString() ));
		    aCurrentExceptionBlock.push( lCatchBlock );
		    catchClauses[i].accept(this);
		    aCurrentExceptionBlock.pop();
		}
		catch( ClassNotFoundException e ){ e.printStackTrace(); }
	    }
    }

    /**
     * visits a try-finally statement
     */
    public void visitTryFinallyStatement(JTryFinallyStatement self,
					 JBlock tryClause,
					 JBlock finallyClause)
    {
	TryBlock lBlock = new TryBlock();
	((ExceptionBlock)aCurrentExceptionBlock.peek()).addElement( lBlock );
	aCurrentExceptionBlock.push( lBlock.getTryBlock() );
	if( tryClause != null ) tryClause.accept( this );
	aCurrentExceptionBlock.pop();

	ExceptionBlock lFinallyBlock = lBlock.getFinallyBlock();
	aCurrentExceptionBlock.push( lFinallyBlock );
	if( finallyClause != null ) finallyClause.accept(this);
	aCurrentExceptionBlock.pop();
    }

    /**
      * visits a throw statement
      * Semantics: adds a throw statement plus any exceptions raised from
      * resolving the expression
      */
    public void visitThrowStatement(JThrowStatement self,
				    JExpression expr)
    {
	try
	{
	    Class lClass = Class.forName( expr.getType().toString() );
	    ((ExceptionBlock)aCurrentExceptionBlock.peek()).addElement( new ExceptionThrow( lClass ));
	}
	catch( ClassNotFoundException e )
	{ e.printStackTrace(); } 
	expr.accept(this);
    }

    /**
     * visits a synchronized statement
     */
    public void visitSynchronizedStatement(JSynchronizedStatement self,
					   JExpression cond,
					   JStatement body)
    {
	if( cond != null ) cond.accept(this);
	if( body != null ) body.accept(this);
    }

    /**
     * visits a switch statement
     */
    public void visitSwitchStatement(JSwitchStatement self,
				     JExpression expr, 
				     JSwitchGroup[] body)
    {
	if( expr != null ) expr.accept(this);
	if( body != null )
	    for( int i = 0; i < body.length; i++ )
		body[i].accept(this);
    }

    /**
     * visits a return statement
     */
  public void visitReturnStatement(JReturnStatement self,
				   JExpression expr)
    {
	if( expr != null ) expr.accept(this);
    }
    
    /**
     * visits a labeled statement
     */
    public void visitLabeledStatement(JLabeledStatement self,
				      String label,
				      JStatement stmt)
    {
	if( stmt != null ) stmt.accept(this);
    }

    /**
     * visits a if statement
     */
    public void visitIfStatement(JIfStatement self,
				 JExpression cond,
				 JStatement thenClause,
				 JStatement elseClause)
    {
	if( cond != null ) cond.accept(this);
	if( thenClause != null ) thenClause.accept(this);
	if( elseClause != null ) elseClause.accept(this);
    }
    
    /**
     * visits a for statement
     */
    public void visitForStatement(JForStatement self,
				  JStatement init,
				  JExpression cond,
				  JStatement incr,
				  JStatement body)
    { 
	if( init != null ) init.accept(this);
	if( cond != null ) cond.accept(this);
	if( incr != null ) incr.accept(this);
	if( body != null ) body.accept(this);
    }
    
    /**
     * visits a compound statement
     */
    public void visitCompoundStatement(JCompoundStatement self,
				       JStatement[] body)
    {
	if( body != null )
	    for( int i = 0; i < body.length; i++ )
		body[i].accept(this);
    }
    
    /**
     * visits an expression statement
     */
    public void visitExpressionStatement(JExpressionStatement self,
					 JExpression expr)
    {
	if( expr != null ) expr.accept(this);
    }

    /**
     * visits an expression list statement
     */
    public void visitExpressionListStatement(JExpressionListStatement self,
					     JExpression[] expr)
    {
	if( expr != null )
	    for( int i = 0; i < expr.length; i++ )
		expr[i].accept(this);
    }

    /**
     * visits a empty statement
     */
    public void visitEmptyStatement(JEmptyStatement self)
    {
    }

    /**
     * visits a do statement
     */
    public void visitDoStatement(JDoStatement self,
				 JExpression cond,
				 JStatement body)
    {
	if( cond != null ) cond.accept(this);
	if( body != null ) body.accept(this);
    }

    /**
     * visits a continue statement
     */
    public void visitContinueStatement(JContinueStatement self,
				       String label)
    {
    }
    
    /**
     * visits a break statement
     */
    public void visitBreakStatement(JBreakStatement self,
				    String label)
    {
    }

    /**
     * visits a block statement
     * 
     */
    public void visitBlockStatement(JBlock self,
				    JStatement[] body,
				    JavaStyleComment[] comments)
    {
	if( body != null )
	    for( int i = 0; i < body.length; i++ )
		body[i].accept(this);
    }
    
    /**
     * visits a type declaration statement
     */
    public void visitTypeDeclarationStatement(JTypeDeclarationStatement self,
				     JTypeDeclaration decl)
    {
	// We dont process inner classes here
	//aNextAnonymous = true;
	//if( decl != null ) decl.accept(this);
    }
    
    // ----------------------------------------------------------------------
    // EXPRESSION
    // ----------------------------------------------------------------------
    
    /**
      * visits an unary expression
      * Semantic: Accepts self.  Polymorphic behavior.
      */
    public void visitUnaryExpression(JUnaryExpression self,
				     int oper,
				     JExpression expr)
    {
	if( expr != null ) expr.accept( this );
    }

    /**
      * visits a type name expression
      * Sematic: Does not produce any exception
      */
    public void visitTypeNameExpression(JTypeNameExpression self,
					CType type)
    {
    }
    
    /**
      * visits a this expression
      * Semantic: Produces the exceptions of the prefix
      */
    public void visitThisExpression(JThisExpression self,
				    JExpression prefix)
    {
	if( prefix != null )
	    prefix.accept( this );
    }
    
    /**
      * visits a super expression
      * Semantic: Does not produce any exception
      */
    public void visitSuperExpression(JSuperExpression self)
    {
    }

    /**
      * visits a shift expression
      * Semantic: Produces the exception of left and right operands.
      */
    public void visitShiftExpression(JShiftExpression self,
				     int oper, 
				     JExpression left, 
				     JExpression right)
    {
	if( left != null ) left.accept( this );
	if( right != null ) right.accept( this );
    }

    /**
      * visits a relational expression
      * Semantic: Produces the exception of left and right operands.
      */
    public void visitRelationalExpression(JRelationalExpression self,
					  int oper, 
					  JExpression left, 
					  JExpression right)
    {
	if( left != null ) left.accept( this );
	if( right != null ) right.accept( this );
    }

    /**
      * visits a prefix expression
      * Semantics: Produces the exceptions of the encapsulated expression
      */
    public void visitPrefixExpression(JPrefixExpression self,
				      int oper, 
				      JExpression expr)
    {
	if( expr != null ) expr.accept( this );
    }
    
    /**
      * visits a postfix expression
      * Semantics: Produces the exceptions of the encapsulated expression
      */
    public void visitPostfixExpression(JPostfixExpression self,
				       int oper, 
				       JExpression expr)
    {
	if( expr != null ) expr.accept( this );
    }

    /**
      * visits a parenthesed expression
      * Semantics: Produces the exceptions of the encapsulated expression
      */
    public void visitParenthesedExpression(JParenthesedExpression self,
					   JExpression expr)
    {
	if( expr != null ) expr.accept( this );
    }

    /**
      * visits an object allocator expression
      * Semantics: Produces the exceptions of the constructor of the object
      * + OutOfMemoryError.
      */
    public void visitNewObjectExpression(JNewObjectExpression self,
					 CClassType type, 
					 JExpression[] params, 
					 JClassDeclaration innerBody)
    {

	// First, generate the OutOfMemoryError
	addEnvironmentException( ExceptionGet.kOutOfMemory );

	// Add the exceptions for evaluating all the args
	if( params != null )
	    for( int i = 0; i < params.length; i++ )
		params[i].accept( this );

	// Add the exceptions for evaluating the constructor for the object
	String lClassName = type.getQualifiedName();
	String lSignature = makeConstructorSignature( self.constructor.getParameters() );
	//if( aJexFiles.contains( lClassName ))
	//{
	//   addAllInternalNonVirtualCallExceptions( lClassName, lSignature );
	//}
	//else
	//{
	    // The class is externally declared
	    addExternalNonVirtualCallExceptions( lClassName, lSignature );
	    //}


	// we dont process inner classes here
    }

    /**
      * visits an array allocator expression
      * Semantics: Produces the exception: NegativeArraySizeException, OutOfMemoryError,
      * plus the initialization exceptions
      */
    public void visitNewArrayExpression(JNewArrayExpression self,
					CType type, 
					JArrayDimsAndInits dims)
    {

	addEnvironmentException( ExceptionGet.kOutOfMemory );
	addEnvironmentException( ExceptionGet.kNegativeArraySize );
	if( dims != null ) dims.accept( this );
    }

    /**
      * visits a name expression
      * Semantic: Adds the exceptions of the prefix
      */
    public void visitNameExpression(JNameExpression self,
				    JExpression prefix, 
				    String ident)
    {
	if( prefix != null ) prefix.accept(this);
    }

    /**
      * visits a binary expression
      * Semantic: Adds the exceptions obtained by evaluating the operands,
      * adds OutOfMemoryError in case of String concatenation, and
      * arithmeticException in case of integer division or remainder.
      */
    public void visitBinaryExpression(JBinaryExpression self,
				      String oper,
				      JExpression left,
				      JExpression right)
    {
	// Check for string concat
	if( oper.equals("+") || oper.equals("+=" ))
	{
	    if( left.getType().toString().equals("java.lang.String") && 
		(right.getType().toString().equals("java.lang.String")))
	    {
		addEnvironmentException( ExceptionGet.kOutOfMemory );
	    }
	}
	if( oper.equals("/") || (oper.equals("%" )))
	{
	    if( (right.getType().toString().equals("int") || 
		 right.getType().toString().equals("short") ||
		 right.getType().toString().equals("long") ||
		 right.getType().toString().equals("byte") ||
		 right.getType().toString().equals("char")) &&
		(left.getType().toString().equals("int") ||
		 left.getType().toString().equals("short") ||
		 left.getType().toString().equals("long") ||
		 left.getType().toString().equals("byte") ||
		 left.getType().toString().equals("char")))
	     {
		 addEnvironmentException( ExceptionGet.kArithmetic );
	     }
	}
	left.accept(this);
	right.accept(this);
    }
    
    /**
     * visits a method call expression
     */
    public void visitMethodCallExpression(JMethodCallExpression self,
					  JExpression prefix,
					  String ident,
					  JExpression[] args)
    {
	// First, generate the NullPointerException
	addEnvironmentException( ExceptionGet.kNullPointer );

	// Add the exceptions for evaluating the prefix
	if( prefix != null ) prefix.accept(this);

	// Add the exceptions for evaluating all the args
	if( args != null )
	    for( int i = 0; i < args.length; i++ )
		args[i].accept(this);

	// Add the exceptions from the actual method call.
	CMethod lMethod = self.method;
	CClass  lClass  = self.method.owner;
	String lClassName = lClass.getQualifiedName();
	String lSignature = makeMethodSignature( lMethod.getIdent(),lMethod.getParameters() );

	if( lMethod.isStatic() )
	{
	    //if( aJexFiles.contains( lClassName ))
	    //{
	    //	addAllInternalNonVirtualCallExceptions( lClassName, lSignature );
	    //}
	    //else
	    //{
		// The class is externally declared
		addExternalNonVirtualCallExceptions( lClassName.replace('/','.'), lSignature );
		//}
	}
	else
	{
	    addVirtualCallExceptions( lClassName.replace('/','.'), lSignature );
	}
    }

    /**
     * visits a local variable expression
     */
    public void visitLocalVariableExpression(JLocalVariableExpression self,
					     String ident)
    {
    }
    
    /**
     * visits an instanceof expression
     */
    public void visitInstanceofExpression(JInstanceofExpression self,
					  JExpression expr, 
					  CType dest)
    {
	if( expr != null ) expr.accept(this);
    }

    /**
     * visits an equality expression
     */
    public void visitEqualityExpression(JEqualityExpression self,
					int oper,
					JExpression left,
					JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

    /**
     * visits a conditional expression
     */
    public void visitConditionalExpression(JConditionalExpression self,
					   JExpression cond,
					   JExpression left,
					   JExpression right)
    {
	if( cond != null ) cond.accept(this);
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

    /**
     * visits a compound expression
     */
    public void visitCompoundAssignmentExpression(JCompoundAssignmentExpression self,
						  int oper,
						  JExpression left,
						  JExpression right)
    {
	if( (left.getType().isArrayType()) && (right.getType().isReference()) )
	    addEnvironmentException( ExceptionGet.kArrayStore );

	left.accept(this);
	right.accept(this);
    }
    
    /**
     * visits a field expression
     */
    public void visitFieldExpression(JClassFieldExpression self,
				     JExpression left, 
				     String ident)
    {
	addEnvironmentException( ExceptionGet.kNullPointer );
	if( left != null ) left.accept(this);
    }
    
    /**
     * visits a class expression
     */
    public void visitClassExpression(JClassExpression self,
				     CType type)
    {
    }

    /**
     * visits a cast expression
     */
    public void visitCastExpression(JCastExpression self,
				    JExpression expr, 
				    CType type)
    {
	addEnvironmentException( ExceptionGet.kClassCast );
	if( expr != null ) expr.accept(this);
    }

    /**
     * visits a cast expression
     */
    public void visitUnaryPromoteExpression(JUnaryPromote self,
					    JExpression expr, 
					    CType type)
    {
	if( expr != null ) expr.accept(this );
    }

    /**
     * visits a compound assignment expression
     */
    public void visitBitwiseExpression(JBitwiseExpression self,
				       int oper,
				       JExpression left,
				       JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }
    
    /**
     * visits an assignment expression
     */
    public void visitAssignmentExpression(JAssignmentExpression self,
					  JExpression left, 
					  JExpression right)
    {
	if( (left.getType().isArrayType()) && (right.getType().isReference()) )
	    addEnvironmentException( ExceptionGet.kArrayStore );

	left.accept(this);
	right.accept(this);
    }
    
    /**
     * visits an array length expression
     */
    public void visitArrayLengthExpression(JArrayLengthExpression self,
					   JExpression prefix)
    {
	addEnvironmentException( ExceptionGet.kNullPointer );
	if( prefix != null ) prefix.accept(this);
    }

    /**
     * visits an array length expression
     */
    public void visitArrayAccessExpression(JArrayAccessExpression self,
					   JExpression prefix, 
					   JExpression accessor)
    {
	addEnvironmentException( ExceptionGet.kNullPointer );
	addEnvironmentException( ExceptionGet.kArrayIndexOutOfBounds );
	if( prefix != null ) prefix.accept(this);
	if( accessor != null ) accessor.accept(this);
    }

  

  // ----------------------------------------------------------------------
  // OTHERS
  // ----------------------------------------------------------------------

  /**
   * visits an array length expression
   */
    public void visitSwitchLabel(JSwitchLabel self,
				 JExpression expr)
    {
	if( expr != null ) expr.accept(this);
    }

    /**
     * visits an array length expression
     */
    public void visitSwitchGroup(JSwitchGroup self,
				 JSwitchLabel[] labels, 
				 JStatement[] stmts)
    {
	if( stmts != null )
	    for( int i = 0; i < stmts.length; i++ )
		stmts[i].accept(this);
    }

    /**
     * visits an array length expression
     */
    public void visitCatchClause(JCatchClause self,
				 JFormalParameter exception, 
				 JBlock body)
    {
	if( exception != null ) exception.accept(this);
	if( body != null ) body.accept(this);
    }
    
  /**
   * visits an array length expression
   */
    public void visitFormalParameters(JFormalParameter self,
				      boolean isFinal, 
				      CType type, 
				      String ident)
    {
    }

  

    /**
     * visits an array initializer expression
     */
    public void visitArrayInitializer(JArrayInitializer self,
				      JExpression[] elems)
    {
	if( elems != null )
	    for( int i = 0; i < elems.length; i++ )
		elems[i].accept(this);
    }

  /**
   * visits an array dimension
   */
  public void visitArrayDimsAndInit(JArrayDimsAndInits self,
			     JExpression[] dims, 
			     JArrayInitializer init)
      {
	  if( dims != null )
	      for( int i = 0; i < dims.length; i++ )
	      {
		  if( dims[i] != null ) dims[i].accept(this);
	      }
	  if( init != null ) init.accept(this);
      }


    /**
     * visits an array length expression
     */
    public void visitExplicitConstructorInvocation(JExplicitConstructorInvocation self,
						   JExpression prefix,
						   String ident,
						   JExpression[] params)
    {
	// Process the prefix
	if( prefix != null ) prefix.accept(this);
	
	// process the parameters
	for( int i = 0; i < params.length; i++ )
	   params[i].accept(this);

	// process the call
	CMethod lMethod = self.method;
	CClass  lClass = lMethod.owner;
	String lClassName = lClass.getQualifiedName();
	String lSignature = makeConstructorSignature( lMethod.getParameters() );

	//if( aJexFiles.contains( lClassName ))
	//{
	//   addAllInternalNonVirtualCallExceptions( lClassName, lSignature );
	//}
	//else
	//{
	    // The class is externally declared
	    addExternalNonVirtualCallExceptions( lClassName, lSignature );
	    //}
    }

    /** -------------------- UNUSED VISIT METHODS -------------------- */

    public void visitInterfaceDeclaration(JInterfaceDeclaration self,
					  int modifiers,
					  String ident,
					  CClassType[] interfaces,
					  JPhylum[] body,
					  JMethodDeclaration[] methods){}

    public void visitComments(JavaStyleComment[] comments){}
    public void visitComment(JavaStyleComment comment){}
    public void visitJavadoc(JavadocComment comment){}
    public void visitClassBody(JTypeDeclaration[] decls,
			       JMethodDeclaration[] methods,
			       JPhylum[] body){}
    public void visitInnerClassDeclaration(JClassDeclaration self,
					   int modifiers,
					   String ident,
					   String superName,
					   CClassType[] interfaces,
					   JTypeDeclaration[] decls,
					   JPhylum[] body,
					   JMethodDeclaration[] methods){}
    public void visitBooleanLiteral(boolean value){}
    public void visitByteLiteral(byte value){}
    public void visitCharLiteral(char value){}
    public void visitDoubleLiteral(double value){}
    public void visitFloatLiteral(float value){}
    public void visitIntLiteral(int value) {}
    public void visitLongLiteral(long value){}
    public void visitShortLiteral(short value){}
    public void visitStringLiteral(String value){}
    public void visitNullLiteral(){}
    public void visitPackageName(String name){}
    public void visitPackageImport(String name){}
    public void visitClassImport(String name){}
}
