/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Originally (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *  
 *  This file was re-written almost entirely for the Jex Analysis Tool
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package at.dms.kjc;

import java.io.File;
import java.io.IOException;
import java.util.Vector;

import at.dms.compiler.Compiler;
import at.dms.compiler.CompilerMessages;
import at.dms.compiler.CWarning;
import at.dms.compiler.InputBuffer;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

import ca.ubc.cs.jex.Workspace;
import ca.ubc.cs.jex.file.ConfigReader;
import ca.ubc.cs.jex.JexFileCollection;
import ca.ubc.cs.jex.Log;
import ca.ubc.cs.jex.util.MethodRelationshipsException;
import ca.ubc.cs.jex.util.PackageNotFoundException;
import ca.ubc.cs.jex.stubs.BasicExtractor;

/**
 * This class implements the entry point of the Java compiler.
 * Revamped to deal with Jex.
 */
public class Main extends Compiler
{
    /** Runs a compilation session
      * @param args the command line arguments
      */
    public boolean run(String[] args)
    {
	System.out.println();
	System.out.println("Jex version 1.2, Copyright (C) 2002 Martin P. Robillard");
	System.out.println("Jex comes with ABSOLUTELY NO WARRANTY" );
	System.out.println("This is free software, and you are welcome to redistribute it");
	System.out.println("under certain conditions; read the GNU General Public License v. 2 for details.");

	// Load the options
	if( !parseArguments(args) )
	    return false;
	
	checkArguments();
	loadConfig();

	if( options.singlepass )
	{
	    System.out.println();
	    System.out.println( "RUNNING IN SINGLE ITERATION MODE" );
	}


	if( options.genstubs )
	    generateStubs();
	    
	if( infiles.size() > 0 )
	{
	    loadWorkspace();
	    parseAndCheck();
	    
	    int lSum = iterate(1);

	    if( options.singlepass )
	    {
		System.out.println( "Checksum: " + lSum );
	    }
	    else
	    {
		int lIteration = 2;
		int lDelta = 0;
	
		do
		{
		    int lNewSum = iterate( lIteration );
		    lDelta = lNewSum - lSum;
		    lSum = lNewSum;
		    lIteration++;
		}
		while( lDelta != 0 );
	    }

	    produceFeedback();
	}
	else
	{
	    System.out.println("No java file was found" );
	    return false;
	}
	
	return true;
    }

    /** Makes sure that the command-line arguments are valid
     */
    private void checkArguments()
    {
	// Make sure the class path is initialized
	initialize();

	// check the validity of the arguments
	if( infiles.size() == 0 )
	{
	    options.usage();
	    inform(KjcMessages.NO_INPUT_FILES);
	    System.exit(0);
	}
	else if( infiles.size() != 1 )
	{
	    options.usage();
	    inform(KjcMessages.WRONG_INPUT );
	    System.exit(0);
	}

	try
	{
	    infiles = verifyFiles(infiles);
	}
	catch( UnpositionedError e )
	{
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}
    }

    /** Initializes the configuration file reader */
    private void loadConfig()
    {
	try
	{
	    aReader = new ConfigReader( (File)infiles.firstElement() );
	    String[] lFiles = aReader.getFiles();
	    infiles.removeAllElements();
	    for( int i = 0; i < lFiles.length ; i++ )
		infiles.addElement( lFiles[i] );
	    
	    infiles = verifyFiles(infiles);
	}
	catch( IOException e )
	{
	    System.out.println( "Error reading config file" );
	    System.exit(0);
	}
	catch( UnpositionedError e )
	{
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}
    }

    /** Generate stubs */
    private void generateStubs()
    {
	try
	{
	    System.out.print( "Generating stubs..." );
	    System.out.flush();
	    long lBegin = System.currentTimeMillis();
		
	    String[] lTargets = aReader.getStubTargets();
	    for( int i = 0; i < lTargets.length; i++ )
	    {
		BasicExtractor lExtractor = new BasicExtractor( lTargets[i] );
		lExtractor.generateJexFiles( new File( aReader.getJexPath() + 
						       File.separator + 
						       lTargets[i].replace('.',File.separatorChar)));
	    }
	    long lEnd = System.currentTimeMillis();
	    aLog.setStubGenTime( lEnd - lBegin );
	    System.out.println( "done (" + lTargets.length + " packages)"  );
	}
	catch( PackageNotFoundException e )
	{
	    System.out.println( "Package not found while generating stubs" );
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}
	catch( IOException e )
	{
	    System.out.println( "IO error while generating stubs" );
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}
	catch( ClassNotFoundException e )
	{
	    System.out.println( "Class not found while generating stubs" );
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}
    }
    
    private void loadWorkspace()
    {
	System.out.print( "Loading workspace..." );
	System.out.flush();
	long lBegin = System.currentTimeMillis();

	try
	{
	    aWorkspace = new Workspace( aReader.getJexPath(), aReader.getPackages() );
	}
	catch( MethodRelationshipsException e )
	{
	    System.out.println( "Cannot properly load workspace" );
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}

	long lEnd = System.currentTimeMillis();
	aLog.setLoadTime( lEnd - lBegin );
	System.out.println( "done" );
    }

    private void parseAndCheck()
    {
	System.out.print( "Parsing and checking..." );
	System.out.flush();
	long lBegin = System.currentTimeMillis();

	aTree = new JCompilationUnit[infiles.size()];

	System.out.println( "Parsing..." );

	for (int count = 0; count < aTree.length; count++)
	{
	    aTree[count] = parseFile((File)infiles.elementAt(count));
	}
	
	if( errorFound ) System.exit(0);
	 
	System.out.println( "Checking interfaces..." );
   
	for( int count = 0; count < aTree.length; count++)
	{
	    checkInterface(aTree[count]);
	}
	
	if(errorFound) System.exit(0);
	    
	System.out.println( "Checking initializers..." );

	for (int count = 0; count < aTree.length; count++)
	{
	    checkInitializers(aTree[count]);
	}
		
	if(errorFound) System.exit(0);
	
	System.out.println( "Checking Bodies..." );

	for( int count = 0; count < aTree.length; count++)
	{
	    checkBody(aTree[count]);
	}

	if(errorFound) 
	{
	    System.out.println( "Error found while compiling.  Exiting..." );
	    System.exit(0);
	}
	long lEnd = System.currentTimeMillis();
	aLog.setParseTime( lEnd - lBegin );
	
	System.out.println("done");
    }

    private int iterate( int pNumber )
    {
	long lBegin = System.currentTimeMillis();
	JexVisitor lVisitor = new JexVisitor( aWorkspace );
	int lSum = 0;
	System.out.print("Iteration " + pNumber + ": " );
	for( int i = 0; i < aTree.length; i++ )
	{
	    System.out.print(".");
	    System.out.flush();
	    aTree[i].accept( lVisitor );
	    JexFileCollection lJexFiles = lVisitor.getJexFiles();
	    try
	    {
		lJexFiles.writeJexFiles( aWorkspace.getJexPath() );
	    }
	    catch( IOException e )
	    {
		System.out.println( "Error writing Jex files: " + e.getMessage() );
		System.exit(0);
	    }
	    aLog.merge( lVisitor.getLog() );
	    lSum += lJexFiles.getCheckSum();
	}
	System.out.println();
	long lEnd = System.currentTimeMillis();
	aLog.addIteration( lEnd - lBegin );
	return lSum;
    }

    private void produceFeedback()
    {
	try
	{
	    aLog.generateReport( "jex-report.txt", infiles, aWorkspace.getJexPath(), aWorkspace.getPackages() );
	    System.out.println();
	    System.out.println( "Jex analysis complete (" + aLog.errors() + " errors, " + aLog.warnings() + " warnings)" ); 
	}
	catch( IOException e )
	{
	    System.out.println( "Error: cannot generate jex report: " + e.getMessage() );
	    System.exit(0);
	}
    }

    /**
     * Parse the argument list
     */
    public boolean parseArguments(String[] args)
    {
	options = new KjcOptions();
	return options.parseCommandLine(args, infiles);
    }

    /**
     * Initialize the compiler (read classpath, check classes.zip)
     */
    protected void initialize()
    {
	if (options.classpath == null)
	{
	    options.classpath = System.getProperty("java.class.path");
	    if (options.classpath == null)
	    {
		options.classpath = ".";
	    }
	}
	at.dms.classfile.ClassPath.init(options.classpath);
	CStdType.init(this);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------
    
    /**
     * @return true if compilation in verbose mode
     */
    public boolean verboseMode()
    { return false; }

    /**
     * @return the optimization level
     */
    public int optimizeCode()
    { return 0; }

    /**
     * Shows an error/warning and set error flag to true
     */
    public KjcPrettyPrinter getPrettyPrinter(String fileName)
    {
	return new KjcPrettyPrinter(fileName);
    }

    /**
     * Entry point
     * @param	args		the command line arguments
     */
    public static void main(String[] args) 
    {
	boolean success = new Main().run(args);
	System.exit(success ? 0 : 1);
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------
    
    /**
     * parse the given file and return a compilation unit
     * side effect: increment error number
     * @param	file		the name of the file (assert exists)
     * @return	the compilation unit defined by this file
     */
    protected JCompilationUnit parseFile(File file)
    {
	InputBuffer buffer;
	
	try
	{
	    buffer = new InputBuffer(file);
	}
	catch( IOException e )
	{
	    inform(e);
	    return null;
	}

	KjcParser		parser; 
	JCompilationUnit	unit;
	long		        lastTime = System.currentTimeMillis();

	parser = new KjcParser(this, buffer);

	try
	{
	    unit = parser.jCompilationUnit();
	}
	catch( antlr.ParserException e )
	{
	    inform(parser.beautifyParseError(e));
	    errorFound = true;
	    unit = null;
	}
	catch( Exception e )
	{
	    //err.println("{" + file.getPath() + ":" + scanner.getLine() + "} " + e.getMessage());
	    e.printStackTrace();
	    errorFound = true;
	    unit = null;
	}

	if( verboseMode())
	{
	    inform(CompilerMessages.FILE_PARSED, file.getPath(), new Long(System.currentTimeMillis() - lastTime));
	}

	buffer.release();
	
	return unit;
    }

    /**
     * check that interface of a given compilation unit is correct
     * side effect: increment error number
     * @param	cunit		the compilation unit
     */
    protected void checkInterface(JCompilationUnit cunit)
    {
	try
	{
	    cunit.checkInterface(this);
	}
	catch( PositionedError e )
	{
	    reportTrouble(e);
	}
    }

    /**
     * check that interface of a given compilation unit is correct
     * side effect: increment error number
     * @param	cunit		the compilation unit
     */
    protected void checkInitializers(JCompilationUnit cunit)
    {
	try
	{
	    cunit.checkInitializers(this, classes);
	}
	catch( PositionedError e )
	{
	    reportTrouble(e);
	}
    }

    /**
     * check that body of a given compilation unit is correct
     * side effect: increment error number
     * @param	cunit		the compilation unit
     */
    protected void checkBody(JCompilationUnit cunit)
    {
	long lastTime = System.currentTimeMillis();
	try
	{
	    cunit.checkBody(this, classes);
	}
	catch( PositionedError e )
	{
	    reportTrouble(e);
	}
	
	if( verboseMode() )
	{
	    inform(CompilerMessages.BODY_CHECKED, cunit.getFileName(), new Long(System.currentTimeMillis() - lastTime));
	}
    }

    // --------------------------------------------------------------------
    // COMPILER
    // --------------------------------------------------------------------
    
    /**
     * Reports a trouble (error or warning).
     *
     * @param	trouble		a description of the trouble to report.
     */
    public void reportTrouble(PositionedError trouble)
    {
	if (trouble instanceof CWarning)
	{
	}
	else
	{
	    if (trouble.getTokenReference() != TokenReference.NO_REF)
	    {
		inform(trouble);
		errorFound = true;
	    }
	}
    }

    /**
     * Returns true iff comments should be parsed (false if to be skipped)
     */
    public boolean parseComments()
    {
	return false;
    }

    /**
     * Returns the classes to generate
     */
    public CSourceClass[] getClasses()
    {
	return (CSourceClass[])at.dms.util.Utils.vectorToArray(classes, CSourceClass.class);
    }

    // ----------------------------------------------------------------------
    // PROTECTED DATA MEMBERS
    // ----------------------------------------------------------------------
    
    protected Vector  infiles = new Vector();
    protected boolean errorFound;

    protected KjcOptions options;

    // all generated classes
    private Vector classes = new Vector(100);

    private Log  aLog = new Log();
    private ConfigReader aReader;
    private Workspace aWorkspace;
    private JCompilationUnit[] aTree;
}
