/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex;

import java.util.Vector;
import java.util.HashSet;
import java.util.Date;
import java.util.Iterator;
import java.text.DateFormat;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;

/** Creates a log of the Jex execution.
  * @version $Revision: 1.2 $ $Date: 2002/03/08 21:56:46 $
  */
public class Log
{
    private HashSet aMethods;   // Methods not found
    private HashSet aJexFiles;  // Jex files not found
    private Vector  aErrors;    // Other errors
    private long    aLoadTime;  // Time to load the workspace
    private long    aParseTime; // Parsing and checking time
    private long    aStubGenTime; // Time to generate the stubs
    private Vector  aIterations; // Vector of Long.
    
    public Log()
    {
	aMethods = new HashSet();
	aJexFiles = new HashSet();
	aErrors = new Vector();
	aIterations = new Vector();
	aStubGenTime = 0;
    }

    public void merge( Log pLog )
    {
	aMethods.addAll(pLog.aMethods);
	aJexFiles.addAll(pLog.aJexFiles);
	aErrors.addAll(pLog.aErrors);
    }

    // Classname.signature 
    public void addMethod( String pMethod )
    {
	aMethods.add( pMethod );
    }

    public void addJexFile( String pClassName )
    {
	aJexFiles.add( pClassName );
    }

    public void addError( String pClass, String pMessage )
    {
	aErrors.addElement( new JexError( pClass, pMessage) );
    }

    public int errors()
    {
	return aJexFiles.size() + aErrors.size();
    }

    public int warnings()
    {
	return aMethods.size();
    }

    public void setLoadTime( long pTime )
    { aLoadTime = pTime; }

    public void setParseTime( long pTime )
    { aParseTime = pTime; }

    public void setStubGenTime( long pTime )
    { aStubGenTime = pTime; }

    public void addIteration( long pTime )
    {
	aIterations.addElement( new Long( pTime ));
    }

    public static String formatTime( long pTime )
    {
	long lHours = pTime / 3600000;
	long lMinutes = (pTime % 3600000)/60000;
	long lSeconds = (pTime % 3600000)%60000/1000;
	long lmSeconds = (pTime % 3600000)%60000%1000;

	String lSHours = "00";
	if( lHours > 0 ) lSHours = new Long(lHours).toString();

	String lSMinutes = "00";
	if( lMinutes > 0 ) lSMinutes = new Long(lMinutes).toString();
	if( lSMinutes.length() == 1) lSMinutes = "0" + lSMinutes;

	String lSSeconds = "00";
	if( lSeconds > 0 ) lSSeconds = new Long(lSeconds).toString();
	if( lSSeconds.length() == 1) lSSeconds = "0" + lSSeconds;

	String lSmSeconds = new Long(lmSeconds).toString();
	if( lSmSeconds.length() == 1 ) lSmSeconds = "000" + lSmSeconds;
	else if( lSmSeconds.length() == 2 ) lSmSeconds = "00" + lSmSeconds;
	else if( lSmSeconds.length() == 3 ) lSmSeconds = "0" + lSmSeconds;

	return lSHours + ":" + lSMinutes + ":" + lSSeconds + "." + lSmSeconds;
    }

    public static void main(String[] args)
    {
	System.out.println( formatTime( Long.parseLong( args[0] )));
    }

    public void generateReport( String pName, Vector pFiles, String pJexPath, String[] pPackages ) throws IOException
    {
	PrintWriter lOut = new PrintWriter( new FileWriter( pName ));
	lOut.println( "Jex analysis report produced on " + 
		      DateFormat.getDateTimeInstance().format(new Date()));
	lOut.println();
	lOut.println( "Jex path: " + pJexPath );
	lOut.println();
	lOut.println( "Packages used for class hierarchy analysis:" );
	lOut.println( "-------------------------------------------" );
	for( int i = 0; i < pPackages.length; i++ )
	    lOut.println( pPackages[i] );
	lOut.println();
	lOut.println( "Java files analyzed:" );
	lOut.println( "--------------------" );
	for( int i = 0; i < pFiles.size(); i++ )
	    lOut.println( pFiles.elementAt(i) );

	lOut.println();
	lOut.println( "Analysis times:" );
	lOut.println( "---------------");
	lOut.println( "Loading the workspace:      " + formatTime(aLoadTime) );
	if( aStubGenTime > 0 )
	    lOut.println( "Generating stubs:           " + formatTime(aStubGenTime));   
	lOut.println( "Parsing and checking files: " + formatTime(aParseTime) );
	for( int i = 0; i < aIterations.size(); i++ )
	    lOut.println( "Analysis iteration " + (i+1) + ":       " + formatTime(((Long)aIterations.elementAt(i)).longValue()) );

	lOut.println();
	lOut.println( "Warnings:" );
	lOut.println( "---------" );
	if( aMethods.size() == 0 )
	    lOut.println( "None" );
	else
	{
	    Iterator lIt = aMethods.iterator();
	    while( lIt.hasNext() )
		lOut.println( "Method " + lIt.next() + " not found in method database." );
	}

	lOut.println();
	lOut.println( "Missing Jex Files:" );
	lOut.println( "------------------" );
	if( aJexFiles.size() == 0)
	    lOut.println("None" );
	else
	{
	    Iterator lIt = aJexFiles.iterator();
	    while( lIt.hasNext() )
		lOut.println( lIt.next() );
	}

	lOut.println();
	lOut.println( "Other errors:" );
	lOut.println( "-------------" );
	if( aErrors.size() == 0 )
	    lOut.println("None");
	else
	{
	    for( int i = 0; i < aErrors.size(); i++ )
		lOut.println( aErrors.elementAt(i) );
	}

	lOut.close();
    }
}
