/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex.file;

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.util.StringTokenizer;

/** Converts various things  to and from the Jex notation.
  * In the Jex notation, method signature are the name of
  * the method (or <init> for constructors), followed by
  * the list of arguments in the following format:
  * - for class types, the fully-qualified class name, 
  * - for primitive names, the name of the primitive
  * - for array types, the bytecode format, eg, [[B,
  * or ;Ljava.lang.String;
  */
public class Convert
{
    /** Primitive type descriptors */
    public static final String kByte = "byte";
    public static final String kChar = "char";
    public static final String kDouble = "double";
    public static final String kFloat = "float";
    public static final String kInt = "int";
    public static final String kLong = "long";
    public static final String kShort = "short";
    public static final String kBoolean = "boolean";
    public static final String kNull = "null";
    public static final String kVoid = "void";
    public static final String kString = "java.lang.String";

    public static String toJexSignature( Method pMethod ) throws ClassNotFoundException
    {
	return pMethod.getName() + getParameterString( pMethod.getParameterTypes() );
    }

    public static String toJexSignature( Constructor pConstructor ) throws ClassNotFoundException
    {
	return "<init>" + getParameterString( pConstructor.getParameterTypes() );
    }

    /** if inner is true, removes the first parameter */
    public static String toJexSignature( Constructor pConstructor, boolean pInner ) throws ClassNotFoundException
    {
	Class[] lNewParams;
	if( pInner )
	{
	    Class[] lParams = pConstructor.getParameterTypes();
	    if( lParams != null )
	    {
		if( lParams.length > 1 )
		{
		    lNewParams = new Class[ lParams.length - 1 ];
		    for( int i = 1; i < lParams.length; i++ )
			lNewParams[i - 1] = lParams[i];
		}
		else
		    lNewParams = new Class[0];
	    }
	    else
		lNewParams = new Class[0];
	}
	else
	    lNewParams = pConstructor.getParameterTypes();
	return "<init>" + getParameterString( lNewParams );
    }


    public static Method toMethod( String pClassName, String pSignature ) throws ClassNotFoundException, NoSuchMethodException
    {
	Class lClass = Class.forName( pClassName );
	StringTokenizer lParser = new StringTokenizer( pSignature, "(), " );
	int lParamsNb = lParser.countTokens() - 1;
	String lIdent = lParser.nextToken();
	Class[] lParams = new Class[lParamsNb];
	int i = 0;
	while( lParser.hasMoreTokens() )
	{
	    lParams[i] = Convert.toClass( lParser.nextToken() );
	    i++;
	}
	return lClass.getDeclaredMethod( lIdent, lParams );
    }

    public static Class toClass( String pType ) throws ClassNotFoundException
    {
	if( pType.equals( kByte ))
	    return byte.class;
	else if( pType.equals( kChar ))
	    return char.class;
	else if( pType.equals( kDouble ))
	    return double.class;
	else if( pType.equals( kFloat ))
	    return float.class;
	else if( pType.equals( kInt ))
	    return int.class;
	else if( pType.equals( kLong ))
	    return long.class;
	else if( pType.equals( kShort ))
	    return short.class;
	else if( pType.equals( kBoolean ))
	    return boolean.class;
	else if( pType.equals( kVoid ))
	    return void.class;
	else if( pType.equals( kNull ))
	    return null;
	else 
	    return Class.forName( pType );
    }

    private static String getParameterString( Class[] pParams )
    {
	String lReturn = "(";

	for( int i = 0; i < pParams.length -1; i++ )
	{
	    String lCurrent = pParams[i].getName();
	    lReturn += lCurrent + ",";
	}
	
	if( pParams.length > 0 )
	    lReturn += pParams[pParams.length - 1].getName();
	lReturn += ")";
	return lReturn;
    }

    public static void main( String[] args ) throws Exception
    {
	//Class lClass = Class.forName( args[0] );

	//Constructor[] lConst = lClass.getDeclaredConstructors();
	//for( int i = 0; i < lConst.length; i++ )
	//  System.out.println( toJexSignature( lConst[i] ));

	//Method[] lMethods = lClass.getDeclaredMethods();
	//for( int i = 0; i < lMethods.length; i++ )
	//    System.out.println( toJexSignature( lMethods[i] ));
	System.out.println( toMethod( args[0], args[1] ));
    }
}
	    
