/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex.file;

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.util.Vector;
import java.util.Set;
import java.util.HashSet;
import java.util.StringTokenizer;
import java.io.FileReader;
import java.io.IOException;
import java.io.File;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.FileNotFoundException;

import java.util.Iterator;

/** Loads a jex file corresponding to a fully qualified name, and
  * provide services to extract information from it. 
  */
public class JexLoader
{
    private JexPath aJexPath;

    /** Creates a JexLoader with a JexPath
      * @param pPath Path to find the .jex files.
      */
    public JexLoader( String pPath )
    {
	aJexPath = new JexPath( pPath );
    }

    public String getJexPath()
    { return aJexPath.getPath(); }

    /** Returns an array of exception types that can be thrown by 
      * a method of a particular type, found only in the Jex file
      * corresponding to the class.
      * @param pMethod   The method to inquiry about.
      * @return A Set of Class types representing exceptions
      */
    public Set getExceptionsFromSingleClass( String pClass, String pSignature )
	throws NoSuchMethodException, FileNotFoundException, ParseException, ClassNotFoundException
    {
	String lName = pClass.replace('.',File.separatorChar );
	Reader lJexFile = new FileReader( aJexPath.makeJexFileName( pClass ));
	
	// Create a parser for the file
	JexParser lJexParser = new JexParser( lJexFile );
	    
	// Parse the file and return the exceptions
	return lJexParser.getExceptions( pSignature );
    }

    /** Same as other, but fetches the infor from a specific file
      * @param pMethod   The method to inquiry about.
      * @return A Vector of Class types representing exceptions
      */
    private Set getExceptionsFromSingleFile( String pFileName, String pSignature )
	throws NoSuchMethodException, FileNotFoundException, ParseException, ClassNotFoundException
    {
	Reader lJexFile = new FileReader( pFileName );
	
	// Create a parser for the file
	JexParser lJexParser = new JexParser( lJexFile );
	    
	// Parse the file and return the exceptions
	return lJexParser.getExceptions( pSignature );
    }

    /*
      * @param pPackage The package to look for anonymous classes
      * @param pClass The name of the supertype to check for overriding
      * @param pSignature The signature to look for.
      */
    public Set getExceptionsFromAnonymousClasses( String pPackage, String pClass, String pSignature )
	throws IOException, ClassNotFoundException, ParseException
    {
	HashSet lReturn = new HashSet();
	String[] lJexFiles = aJexPath.getAnonymousJexFiles( pPackage );
	
	for( int i = 0; i < lJexFiles.length; i++ )
	{
	    String lFileName = aJexPath.makeJexFileName( pPackage, lJexFiles[i] );
	    String[] lTypes = getTypes( lFileName );
	    if( lTypes != null )
	    {
		boolean doit = false;
		for( int j = 0; j < lTypes.length; j++ )
		{
		    Class lClass = Class.forName( lTypes[j] );
		    if( Class.forName( pClass ).isAssignableFrom( lClass ))
		    {
			doit = true;
			break;
		    }
		}
		if( doit )
		{
		    try
		    {
			Set lExceptions = getExceptionsFromSingleFile( lFileName, pSignature );
			lReturn.addAll( lExceptions );
		    }
		    catch( NoSuchMethodException e )
		    {
			// We postulate that the anonymous class simply did not implement
			// the method.  Do nothing.
		    }
		}
	    }
	}
	return lReturn;
    }
    
    /** null if not anonymous */
    private String[] getTypes( String pJexFileName ) throws FileNotFoundException, IOException
    {
	BufferedReader lFile = new BufferedReader( new FileReader( pJexFileName ));
	String lPackages = lFile.readLine();
	lFile.close();
	if( lPackages == null )
	    return null;
	
	if( lPackages.startsWith( "*anonymous:" ));
	StringTokenizer lParser = new StringTokenizer( lPackages.substring( 11 ), ",; " );
	String[] lReturn = new String[ lParser.countTokens() ];
	int i = 0;
	while( lParser.hasMoreTokens() )
	{
	    lReturn[i] = lParser.nextToken();
	    i++;
	}
	return lReturn;
    }

    public static void main( String[] args ) throws Exception
    {
	JexLoader lLoader = new JexLoader( "c:\\jexfiles" );
        //Class lClass = Class.forName( args[0] + "." + args[1]);
	//Set lExs = null;

	//Method[] lMethods = lClass.getDeclaredMethods();
	//Method lMethod = lMethods[Integer.parseInt( args[2] )];
	//lExs = lLoader.getExceptionsFromAnonymousClasses( args[0], args[1],Convert.toJexSignature( lMethod ));
	Set lExs = lLoader.getExceptionsFromSingleClass( "ca/ubc/cs/se/visualization/visualizer/IntegerHistogram$StackedIntegerBar",
							 "<init>(int)" );

	Iterator lIt = lExs.iterator();
	while( lIt.hasNext() )
	    System.out.println( lIt.next() );
    }
}	    
