/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex.stubs;

import ca.ubc.cs.jex.util.PackageNotFoundException;
import ca.ubc.cs.jex.util.Package;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import ca.ubc.cs.jex.file.Convert;

/** This class produces a jex file for all the classes of a package,
  * with every method only raising the exceptions it declares.
  * @author $Author: mrobilla $
  * @version $Revision: 1.2 $ $Date: 2002/03/08 21:56:06 $
  */
public class BasicExtractor
{
    private Package aPackage;

    /** Creates a new stub extractor for the specified package
      * @param pPackage A package to extract
      */
    public BasicExtractor( Package pPackage )
    { aPackage = pPackage; }

    public BasicExtractor( String pPackage ) throws PackageNotFoundException
    {
	aPackage = new Package( pPackage );
    }

    public void generateJexFiles( File pDestination ) throws IOException, ClassNotFoundException
    {
	String   lPackageName = aPackage.getName();
	String[] lClassNames = aPackage.getClasses();
	Class[] lClasses = getClasses( lPackageName, lClassNames );

	pDestination.mkdirs();

	for( int i = 0; i < lClasses.length; i++ )
	{
	    // Here it would be nice not to generate the stubs for interfaces classes.
	    // However, this must be done for now because Jex relies on the interface
	    // stubs for information about polymorphic calls for which the target method
	    // is not in the method database
	    PrintWriter lOutput = new PrintWriter( new FileWriter( pDestination.getAbsolutePath() + 
								   File.separator + lClassNames[i] + ".jex" ));	

	    Constructor[] lCons = lClasses[i].getDeclaredConstructors();
	    
	    boolean lDefault = false;
	    for( int j = 0; j < lCons.length; j++ )
	    {
		if( lCons[j].getParameterTypes().length < 2 )
		    lDefault = true;

		// HACK: Because it's not possible to tell whether a class is
		// an anonymous class or not, and because anonymous classes
		// have an extra parameter for each of their constructor,
		// as a preventative measure, we add a constructor with
		// the first parameter removed corresponding to each generated
		// parameter.
		lOutput.println( Convert.toJexSignature( lCons[j] ));
		lOutput.println("{");
		
		Class[] lExceptions = lCons[j].getExceptionTypes();
		for( int k = 0; k < lExceptions.length; k++ )
		    lOutput.println( "   throws " + lExceptions[k].getName() + ";");
		
		lOutput.println( "}" );
		
		lOutput.println( Convert.toJexSignature( lCons[j], true ));
		lOutput.println("{");
		for( int k = 0; k < lExceptions.length; k++ )
		    lOutput.println( "   throws " + lExceptions[k].getName() + ";");
		
		lOutput.println( "}" );
	    }
	    if( lDefault == false )
	    {
		lOutput.println( "<init>()" );
		lOutput.println("{");
		lOutput.println("}");
	    }

	    Method[] lMethods = lClasses[i].getDeclaredMethods();
	    for( int j = 0; j < lMethods.length; j++ )
	    {
		lOutput.println( Convert.toJexSignature( lMethods[j] ));
		lOutput.println( "{" );
		
		Class[] lExceptions = lMethods[j].getExceptionTypes();
		for( int k = 0; k < lExceptions.length; k++ )
		    lOutput.println( "   throws " + lExceptions[k].getName() + ";");
		
		lOutput.println( "}" );
	    }
	    lOutput.close();
	}
    }

    public static Class[] getClasses( String pPackageName, String[] pClasses ) throws ClassNotFoundException
    {
	Class[] lReturn = new Class[ pClasses.length ];
	for( int i = 0 ; i < pClasses.length; i++ )
	{
	    lReturn[i] = Class.forName( pPackageName + "." + pClasses[i] );
	    
	}
	return lReturn;
    }
    
    public static void main( String[] args )
    {
	if( args.length != 2 )
	{
	    usage();
	    System.exit(0);
	}
	
	try
	{
	    header();
	    Package lPackage = new Package( args[0] );
	    System.out.println( "Generating Jex files for package " + lPackage.getName() + 
				" ("+ lPackage.getClasses().length + " classes)...");
	    BasicExtractor lExtractor = new BasicExtractor( lPackage );
	    lExtractor.generateJexFiles( new File( args[1] ) );
	    System.out.println( "File generation completed successfully" );
	}
	catch( PackageNotFoundException e )
	{
	    System.out.println( "Error: Could not find package " + args[0] + " on the current system" );
	    System.exit(0);
	}
	catch( IOException e )
	{
	    System.out.println( "Error: IOException while generating Jex files" );
	    System.out.println( e.getMessage() );
	    System.exit(0);
	}
	catch( ClassNotFoundException e )
	{
	    System.out.println( "Error Trying to load unexisting class: " + e.getMessage() );
	    System.exit(0);
	}
    }

    private static void header()
    {
	System.out.println();
	System.out.println("Jex Basic Extractor version 1.0, Copyright (C) 2001 Martin P. Robillard");
	System.out.println("Jex Basic Extractor comes with ABSOLUTELY NO WARRANTY" );
	System.out.println("This is free software, and you are welcome to redistribute it");
	System.out.println("under certain conditions; read the GNU General Public License v. 2 for details.");
	System.out.println();
    }

    private static void usage()
    {
	header();
	System.out.println( "Usage:" );
	System.out.println();
	System.out.println( "Jex Basic Extractor generates Jex stub files for all the");
	System.out.println( "classes of a Java package.  The stubs provide only information");
	System.out.println( "about the exceptions declared in a method's signature.");
	System.out.println();
	System.out.println( "To execute Jex Basic Extractor, execute:" );
	System.out.println( "java jex.bytecode.BasicExtractor <Package name> <destination>, where");
	System.out.println( "<package name> is the fully-qualified name of the package, and" );
	System.out.println( "<destination>  is directory that should contain the Jex stubs." );
	System.exit(0);
    }
}
