/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex.util;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

/** A BasicMethodRelationships implements MethodRelationships
  * using a huge hash table of method.  Every method is
  * a key in the hash table.  The corresponding value is the
  * list of kid methods, or null if the list was not computed.
  * Adding a package causes all the relationships to be invalidated,
  * so it's better to have an initialization phases where 
  * packages are added and a query phase where the database
  * is queried.
  * @author $Author: mrobilla $
  * @version $Revision: 1.2 $ $Date: 2002/03/08 21:56:21 $
  */
public class BasicMethodRelationships implements MethodRelationships
{
    private Hashtable aRelationships = new Hashtable();
    private final static Integer kMarker = new Integer(0);
    private boolean aQueried = false;

    /** Adds all the methods of all the classes in the
      * package to the relationship database.
      * @param pPackage The fully qualified name of the
      * package to add.
      * @exception MethodRelationshipException if the
      * package cannot be loaded.
      */
    public void addPackage( String pPackage ) throws MethodRelationshipsException
    {
	if( aQueried )
	    invalidateAll();

	try
	{
	    Package lPackage = new Package( pPackage );
	    String[] lClasses = lPackage.getClasses();

	    for( int i = 0; i < lClasses.length; i++ )
	    {
		Class lClass = Class.forName( lPackage.getName() + "." + lClasses[i] );
		Method[] lMethods = lClass.getDeclaredMethods();
		for( int j = 0; j < lMethods.length; j++ )
		    aRelationships.put( lMethods[j], kMarker );
	    }
	}
	catch( PackageNotFoundException e )
	{
	    throw new MethodRelationshipsException( "Cannot add package " + 
						    pPackage + ": " + e.getMessage() );
	}
	catch( ClassNotFoundException e )
        {
	    throw new MethodRelationshipsException( "Error loading package " + pPackage );
	}
    }

    /** Deletes all cached method relationships computations.
     */
    private void invalidateAll()
    {
	Enumeration lKeys = aRelationships.keys();
	while( lKeys.hasMoreElements() )
	{
	    Object lKey = lKeys.nextElement();
	    aRelationships.put( lKey, kMarker );
	}
	aQueried = false;
    }

    /** Returns all the methods that directly or indirectly
      * implement or override the argument method.
      * This follows article 8.4.6.1 of the Java Language
      * Specifications, with the exception that the second 
      * point of item 2 is not implemented, so that a method m1 can
      * override m2 simply m2 is non-private (and other restrictions apply).
      * this approach is over-conservative.
      * @param pMethod The parent method.
      * @return A vector of method ovverriding or implementing
      * the argument method.
      * @exception MethodRelationshipException if the argument
      * method cannot be found in the database.
      */
    public Vector getImplementations( Method pMethod ) throws MethodRelationshipsException
    {
	Object lValue = aRelationships.get( pMethod );
	if( lValue == null )
	    throw new MethodRelationshipsException( "Method " + pMethod.toString() + 
						   " not found in the database");
	else
	{
	    aQueried = true;
	    if( lValue instanceof Integer )
	    {
		Vector lKids = computeRelationships( pMethod );
		aRelationships.put( pMethod, lKids );
		return lKids;
	    }
	    else
		return (Vector)lValue;
	}
    }

    /** Determines all the possible implementations of
	the pMethod signature.  Implements 8.4.6.1 of JLS 
        Note, final is not checked here because checked by
        the compiler
    */
    private Vector computeRelationships( Method pMethod )
    {
	aQueried = false;
	Vector  lReturn = new Vector();
	
	Class   lHost = pMethod.getDeclaringClass();

	// If the method is private, it can't be overriden or implemented
	boolean lPrivate = Modifier.isPrivate( pMethod.getModifiers() );

	if( lPrivate )
	{
	    return lReturn;
	}
	    
	Enumeration lKeys = aRelationships.keys();
	while( lKeys.hasMoreElements() )
	{
	    Method lMethod = (Method)lKeys.nextElement();
	    Class lCHost = lMethod.getDeclaringClass();

	    // Check if host classes are compatible
	    if( lHost.isAssignableFrom( lCHost ))
	    {
		// Check if the signature are equal
		if( signatureEqual( pMethod, lMethod ))
		{
		    boolean lCAbstract = Modifier.isAbstract( lMethod.getModifiers() );

		    // Check if current method has a body
		    if( (lCHost.isInterface() != true ) &&
			(lCAbstract != true ))
		    {
			lReturn.addElement( lMethod );
		    }
		}
	    }
	}
	return lReturn;
    }
	
    /** Returns true if the two methods have equal
     * identifiers, return types, and parameter types
     */
    private static boolean signatureEqual( Method pMethod1, Method pMethod2 )
    {
	if( pMethod1.getName().equals( pMethod2.getName()) == false )
	    return false;

	if( pMethod1.getReturnType().equals( pMethod2.getReturnType() ) == false )
	    return false;

	Class[] lParams1 = pMethod1.getParameterTypes();
	Class[] lParams2 = pMethod2.getParameterTypes();
	
	if( lParams1.length != lParams2.length )
	    return false;

	for( int i = 0; i < lParams1.length; i++ )
	{
	    if( lParams1[i].equals( lParams2[i] ) == false )
		return false;
	}

	return true;
    }

    /** For testing purposes
     */
    public static void main( String[] args ) throws Exception
    {
	MethodRelationships lDB = new BasicMethodRelationships();
	lDB.addPackage( "concerns" );
	Class lClass = Class.forName( "concerns.TestB" );
	Method lMethod = lClass.getDeclaredMethods()[0];
	System.out.println( "Testing method " + lMethod );
	Vector lImpls = lDB.getImplementations( lMethod );
	for( int i = 0; i < lImpls.size(); i++ )
	    System.out.println( lImpls.elementAt( i ));
    }
	
}	    
