/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex.util;

import java.io.File;
import java.io.Reader;
import java.io.IOException;
import java.io.FileNotFoundException;

/** A PackageAccessor represent an abstract object that can serve files
  * containing the class files for a package.
  * It can be either a directory or an archive file.
  * @author $Author: mrobilla $
  * @version $Revision: 1.2 $ $Date: 2002/03/08 21:56:23 $
  */
public abstract class PackageAccessor
{
    /** File representing the accessor */
    protected File aAccessor;

    /** Stores the value in aAccessor.  This constructor should only be invoked
      * by constructors of subtypes.
      * @param pPath Path corresponding to the accessor, either a 
      * directory or a valid archive format.
      */
    protected PackageAccessor( File pPath )
    {
	aAccessor = pPath;
    }

    /** Official constructor for this class.  Picks the right subclass and
      * creates a specialized PackageAccessor.
      * @param pFile A File object representing the location of the package
      * on the system.
      * @exception UnknownArchiveFormatException If pFile does not represent 
      * a file of known format.
      * @exception FileNotFoundException If pFile cannot be located.
      */
    public static final PackageAccessor createPackageAccessor( File pFile )
	throws IOException
    {

	if( pFile.exists() )
	{
	    if( DirectoryPackageAccessor.match( pFile ) )
		return new DirectoryPackageAccessor( pFile );
	    else if( ZipPackageAccessor.match( pFile ))
		return new ZipPackageAccessor( pFile );
	    else if( JarPackageAccessor.match( pFile ))
		return new JarPackageAccessor( pFile );
	    else
		throw new UnknownArchiveFormatException( pFile );
	}
	else
	{
	    throw new FileNotFoundException(pFile.getAbsolutePath());
	}
    }

    /** Returns true if the argument can be correctly represented by the
      * subclass.  This method acts a an interface and has to be redefined satically
      * in all subclasses.
      * @param pFile File to check if we can match to.
      * @return true if the argument is a valid directory
      */
    public static boolean match( File pFile )
    {
	throw new IllegalAccessError( "Method match(File) not defined in a subclass of PackageAccessor");
    }

    /** Returns the names of the classes contained in the package specified as argument,
      * @param pPackageName Fully qualified package name, e.g., "java.lang".
      * @return An array of strings containing the classes in the package.  Null if the
      * package cannot be found in the current structure.
      * @exception IOException if it is impossible to open the
      * specified file.
      */
    public abstract String[] getClasses( String pPackageName ) throws IOException;

    /** Takes a path and a file name and create an absolute filename, 
     * ensuring that the separator exists between the path and the filename */
    protected static String createAbsoluteFileName( File pPath, String pFileName )
    {
	String lReturn = pPath.getAbsolutePath();
	if( lReturn.charAt( lReturn.length() - 1 ) != File.separatorChar )
	{
	    lReturn = lReturn + File.separator;
	}
	
	lReturn = lReturn + pFileName;
	return lReturn;
    }
}
